import bloqade.ir.location as location
import bloqade.ir.control.sequence as sequence
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.field as field
import bloqade.ir.control.waveform as waveform
import bloqade.ir.scalar as scalar
import bloqade.ir.analog_circuit as analog_circuit

from bloqade.ir.visitor.analog_circuit import (
    # AnalogCircuitVisitor,
    AnalogCircuitVisitor,
)
from bloqade.ir.visitor.location import LocationVisitor
from bloqade.ir.visitor.waveform import WaveformVisitor
from bloqade.ir.visitor.scalar import ScalarVisitor

import numbers
from typing import Any, Dict, Union
from decimal import Decimal


class AssignScalar(ScalarVisitor):
    def __init__(self, mapping: Dict[str, Decimal]):
        self.mapping = dict(mapping)

    def visit_literal(self, ast: scalar.Literal) -> scalar.Scalar:
        return ast

    def visit_variable(
        self, ast: scalar.Variable
    ) -> Union[scalar.Literal, scalar.Variable]:
        if ast.name in self.mapping:
            return scalar.AssignedVariable(ast.name, self.mapping[ast.name])

        return ast

    def visit_assigned_variable(
        self, ast: scalar.Variable
    ) -> Union[scalar.Literal, scalar.AssignedVariable]:
        if ast.name in self.mapping:
            raise ValueError(f"Variable {ast.name} already assigned to {ast.value}.")

        return ast

    def visit_negative(self, ast: scalar.Negative) -> scalar.Scalar:
        return -self.visit(ast.expr)

    def visit_add(self, ast: scalar.Add) -> scalar.Scalar:
        return self.visit(ast.lhs) + self.visit(ast.rhs)

    def visit_div(self, ast: scalar.Div) -> scalar.Scalar:
        return self.visit(ast.lhs) / self.visit(ast.rhs)

    def visit_mul(self, ast: scalar.Mul) -> scalar.Scalar:
        return self.visit(ast.lhs) * self.visit(ast.rhs)

    def visit_min(self, ast: scalar.Min) -> scalar.Scalar:
        return scalar.Scalar.canonicalize(
            scalar.Min(frozenset(list(map(self.visit, ast.scalars))))
        )

    def visit_max(self, ast: scalar.Min) -> scalar.Scalar:
        return scalar.Scalar.canonicalize(
            scalar.Max(frozenset(list(map(self.visit, ast.scalars))))
        )

    def emit(self, ast: scalar.Scalar) -> scalar.Scalar:
        # canonicalize at the end
        return scalar.Scalar.canonicalize(self.visit(ast))


class AssignWaveform(WaveformVisitor):
    def __init__(self, mapping: Dict[str, numbers.Real]):
        self.scalar_visitor = AssignScalar(mapping)
        self.mapping = dict(mapping)

    def visit_constant(self, ast: waveform.Constant) -> Any:
        value = self.scalar_visitor.emit(ast.value)
        duration = self.scalar_visitor.emit(ast.duration)
        return waveform.Constant(value, duration)

    def visit_linear(self, ast: waveform.Linear) -> Any:
        start = self.scalar_visitor.emit(ast.start)
        stop = self.scalar_visitor.emit(ast.stop)
        duration = self.scalar_visitor.emit(ast.duration)
        return waveform.Linear(start, stop, duration)

    def visit_poly(self, ast: waveform.Poly) -> Any:
        checkpoints = list(map(self.scalar_visitor.emit, ast.coeffs))
        duration = self.scalar_visitor.emit(ast.duration)
        return waveform.Poly(checkpoints, duration)

    def visit_python_fn(self, ast: waveform.PythonFn) -> Any:
        duration = self.scalar_visitor.emit(ast.duration)
        new_parameters = []
        default_param_values = dict(ast.default_param_values)

        new_parameters = list(map(self.scalar_visitor.emit, ast.parameters))

        # remove default parameters that are overwritten by the assignment
        for param in new_parameters:
            if (
                isinstance(param, scalar.AssignedVariable)
                and param.name in default_param_values
            ):
                default_param_values.pop(param.name)

        return waveform.PythonFn(ast.fn, duration, new_parameters, default_param_values)

    def visit_add(self, ast: waveform.Add) -> Any:
        return waveform.Add(self.visit(ast.left), self.visit(ast.right))

    def visit_alligned(self, ast: waveform.AlignedWaveform) -> Any:
        if isinstance(ast.value, scalar.Scalar):
            value = self.scalar_visitor.emit(ast.value)
        else:
            value = ast.value

        return waveform.AlignedWaveform(self.visit(ast.waveform), ast.alignment, value)

    def visit_append(self, ast: waveform.Append) -> Any:
        return waveform.Append(list(map(self.visit, ast.waveforms)))

    def visit_smooth(self, ast: waveform.Smooth) -> Any:
        static_radius = self.scalar_visitor.emit(ast.radius)
        return waveform.Smooth(static_radius, ast.kernel, self.visit(ast.waveform))

    def visit_negative(self, ast: waveform.Negative) -> Any:
        return waveform.Negative(self.visit(ast.waveform))

    def visit_record(self, ast: waveform.Record) -> Any:
        if ast.var.name in self.mapping:
            return self.visit(ast.waveform)
        else:
            return waveform.Record(self.visit(ast.waveform), ast.var)

    def visit_sample(self, ast: waveform.Sample) -> Any:
        dt = self.scalar_visitor.emit(ast.dt)
        return waveform.Sample(self.visit(ast.waveform), ast.interpolation, dt)

    def visit_scale(self, ast: waveform.Scale) -> Any:
        scale = self.scalar_visitor.emit(ast.scalar)
        return waveform.Scale(scale, self.visit(ast.waveform))

    def visit_slice(self, ast: waveform.Slice) -> Any:
        start = ast.interval.start
        stop = ast.interval.stop

        if start is not None:
            start = self.scalar_visitor.emit(start)

        if stop is not None:
            stop = self.scalar_visitor.emit(stop)

        return waveform.Slice(self.visit(ast.waveform), scalar.Interval(start, stop))

    def emit(self, ast: waveform.Waveform) -> waveform.Waveform:
        return waveform.Waveform.canonicalize(self.visit(ast))


class AssignLocation(LocationVisitor):
    def __init__(self, mapping: Dict[str, numbers.Real]):
        self.mapping = dict(mapping)
        self.scalar_visitor = AssignScalar(mapping)

    def visit_chain(self, ast: location.Chain) -> location.Chain:
        return location.Chain(*ast.shape, self.scalar_visitor.emit(ast.lattice_spacing))

    def visit_square(self, ast: location.Square) -> location.Square:
        return location.Square(
            *ast.shape, self.scalar_visitor.emit(ast.lattice_spacing)
        )

    def visit_rectangular(self, ast: location.Rectangular) -> location.Rectangular:
        return location.Rectangular(
            ast.shape[0],
            ast.shape[1],
            self.scalar_visitor.emit(ast.lattice_spacing_x),
            self.scalar_visitor.emit(ast.lattice_spacing_y),
        )

    def visit_triangular(self, ast: location.Triangular) -> location.Triangular:
        return location.Triangular(
            *ast.shape, self.scalar_visitor.emit(ast.lattice_spacing)
        )

    def visit_honeycomb(self, ast: location.Honeycomb) -> location.Honeycomb:
        return location.Honeycomb(
            *ast.shape, self.scalar_visitor.emit(ast.lattice_spacing)
        )

    def visit_kagome(self, ast: location.Kagome) -> location.Kagome:
        return location.Kagome(
            *ast.shape, self.scalar_visitor.emit(ast.lattice_spacing)
        )

    def visit_lieb(self, ast: location.Lieb) -> location.Lieb:
        return location.Lieb(*ast.shape, self.scalar_visitor.emit(ast.lattice_spacing))

    def visit_list_of_locations(
        self, ast: location.ListOfLocations
    ) -> location.ListOfLocations:
        return location.ListOfLocations(list(map(self.visit, ast.location_list)))

    def visit_location_info(self, ast: location.LocationInfo) -> location.LocationInfo:
        return location.LocationInfo(
            tuple(map(self.scalar_visitor.emit, ast.position)),
            bool(ast.filling.value),
        )

    def visit_parallel_register(
        self, ast: location.ParallelRegister
    ) -> location.ParallelRegister:
        return location.ParallelRegister(
            self.visit(ast._register), self.scalar_visitor.emit(ast._cluster_spacing)
        )


class AssignAnalogCircuit(AnalogCircuitVisitor):
    def __init__(self, mapping: Dict[str, numbers.Real]):
        self.waveform_visitor = AssignWaveform(mapping)
        self.scalar_visitor = AssignScalar(mapping)
        self.location_visitor = AssignLocation(mapping)
        self.mapping = dict(mapping)

    def visit_analog_circuit(self, ast: analog_circuit.AnalogCircuit) -> Any:
        return analog_circuit.AnalogCircuit(
            self.visit(ast.register), self.visit(ast.sequence)
        )

    def visit_register(self, ast: location.AtomArrangement) -> location.AtomArrangement:
        return self.location_visitor.visit(ast)

    def visit_parallel_register(
        self, ast: location.ParallelRegister
    ) -> location.ParallelRegister:
        return self.location_visitor.visit(ast)

    def visit_sequence(self, ast: sequence.Sequence) -> sequence.Sequence:
        return sequence.Sequence(
            {
                coupling_name: self.visit(sub_pulse)
                for coupling_name, sub_pulse in ast.pulses.items()
            }
        )

    def visit_append_sequence(self, ast: sequence.Append) -> sequence.Append:
        return sequence.Append(list(map(self.visit, ast.sequences)))

    def visit_slice_sequence(self, ast: sequence.Slice) -> sequence.Slice:
        return sequence.Slice(
            self.visit(ast.sequence), self.scalar_visitor.emit(ast.interval)
        )

    def visit_named_sequence(
        self, ast: sequence.NamedSequence
    ) -> sequence.NamedSequence:
        return sequence.NamedSequence(self.visit(ast.sequence), ast.name)

    def visit_pulse(self, ast: pulse.Pulse) -> pulse.Pulse:
        return pulse.Pulse(
            {
                field_name: self.visit(sub_field)
                for field_name, sub_field in ast.fields.items()
            }
        )

    def visit_append_pulse(self, ast: pulse.Append) -> pulse.Append:
        return pulse.Append(list(map(self.visit, ast.pulses)))

    def visit_slice_pulse(self, ast: pulse.Slice) -> pulse.Slice:
        return pulse.Slice(self.visit(ast.pulse), self.scalar_visitor(ast.interval))

    def visit_named_pulse(self, ast: pulse.NamedPulse) -> Any:
        return pulse.NamedPulse(ast.name, self.visit(ast.pulse))

    def visit_field(self, ast: field.Field) -> field.Field:
        return field.Field(
            {self.visit(sm): self.visit(wf) for sm, wf in ast.drives.items()}
        )

    def visit_uniform_modulation(self, ast: field.UniformModulation) -> Any:
        return ast

    def visit_location(self, ast: field.Location) -> Any:
        return ast

    def visit_scaled_locations(self, ast: field.ScaledLocations) -> Any:
        return field.ScaledLocations(
            {
                self.visit(loc): self.scalar_visitor.emit(scale)
                for loc, scale in ast.value.items()
            }
        )

    def visit_run_time_vector(
        self, ast: field.RunTimeVector
    ) -> Union[field.RunTimeVector, field.AssignedRunTimeVector]:
        if ast.name in self.mapping:
            return field.AssignedRunTimeVector(ast.name, self.mapping[ast.name])

        return ast

    def visit_assigned_run_time_vector(
        self, ast: field.AssignedRunTimeVector
    ) -> field.AssignedRunTimeVector:
        if ast.name in self.mapping:
            raise ValueError(f"Variable {ast.name} already assigned to {ast.value}.")

        return ast

    def visit_waveform(self, ast: waveform.Waveform) -> waveform.Waveform:
        return self.waveform_visitor.emit(ast)

    def emit(self, ast: analog_circuit.AnalogCircuit) -> analog_circuit.AnalogCircuit:
        return self.visit(ast)



from pydantic import BaseModel
from decimal import Decimal

__all__ = ["QuEraCapabilities"]


class RydbergGlobalCapabilities(BaseModel):
    rabi_frequency_min: Decimal
    rabi_frequency_max: Decimal
    rabi_frequency_resolution: Decimal
    rabi_frequency_slew_rate_max: Decimal
    detuning_min: Decimal
    detuning_max: Decimal
    detuning_resolution: Decimal
    detuning_slew_rate_max: Decimal
    phase_min: Decimal
    phase_max: Decimal
    phase_resolution: Decimal
    time_min: Decimal
    time_max: Decimal
    time_resolution: Decimal
    time_delta_min: Decimal


class RydbergLocalCapabilities(BaseModel):
    detuning_min: Decimal
    detuning_max: Decimal
    detuning_slew_rate_max: Decimal
    site_coefficient_min: Decimal
    site_coefficient_max: Decimal
    number_local_detuning_sites: int
    spacing_radial_min: Decimal
    time_resolution: Decimal
    time_delta_min: Decimal


class RydbergCapabilities(BaseModel):
    c6_coefficient: Decimal
    global_: RydbergGlobalCapabilities
    local: RydbergLocalCapabilities

    class Config:
        allow_population_by_field_name = True
        fields = {"global_": "global"}


class LatticeGeometryCapabilities(BaseModel):
    spacing_radial_min: Decimal
    spacing_vertical_min: Decimal
    position_resolution: Decimal
    number_sites_max: int


class LatticeAreaCapabilities(BaseModel):
    width: Decimal
    height: Decimal


class LatticeCapabilities(BaseModel):
    number_qubits_max: int
    area: LatticeAreaCapabilities
    geometry: LatticeGeometryCapabilities


class TaskCapabilities(BaseModel):
    number_shots_min: int
    number_shots_max: int


class DeviceCapabilities(BaseModel):
    task: TaskCapabilities
    lattice: LatticeCapabilities
    rydberg: RydbergCapabilities


class QuEraCapabilities(BaseModel):
    version: str
    capabilities: DeviceCapabilities



import bloqade.ir.control.waveform as waveform
import bloqade.ir.control.field as field
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.sequence as sequence
import bloqade.ir.analog_circuit as analog_circuit
from bloqade.ir.visitor.waveform import WaveformVisitor
from bloqade.ir.visitor.analog_circuit import AnalogCircuitVisitor
from decimal import Decimal
from beartype.typing import Any, Dict
from beartype import beartype
from pydantic.dataclasses import dataclass


@dataclass(frozen=True)
class IsConstantWaveformResult:
    is_constant: bool
    constant_waveform: waveform.Constant


class IsConstantWaveform(WaveformVisitor):
    @beartype
    def __init__(self, assignments: Dict[str, Decimal]) -> None:
        self.assignments = dict(assignments)
        self.is_constant = True

    def visit_constant(self, _: waveform.Constant) -> Any:
        pass

    def visit_linear(self, ast: waveform.Linear) -> Any:
        diff = ast.stop(**self.assignments) - ast.start(**self.assignments)
        self.is_constant = self.is_constant and (diff == 0)

    def visit_poly(self, ast: waveform.Poly) -> Any:
        coeffs = [coeff(**self.assignments) for coeff in ast.coeffs]
        if any(coeff != 0 for coeff in coeffs[1:]):
            self.is_constant = False

    def visit_python_fn(self, ast: waveform.PythonFn) -> Any:
        # can't analyze python functions, assume it's not constant
        self.is_constant = False

    def visit_append(self, ast: waveform.Append) -> Any:
        value = None
        for wf in ast.waveforms:
            result = IsConstantWaveform(self.assignments).emit(wf)
            if value is None:
                value = result.constant_waveform.value

            self.is_constant = (self.is_constant and result.is_constant) and (
                result.constant_waveform.value == value
            )

            if not self.is_constant:
                return

    def visit_add(self, ast: waveform.Add) -> Any:
        left_duration = ast.left.duration(**self.assignments)
        right_duration = ast.right.duration(**self.assignments)

        if left_duration != right_duration:
            self.is_constant = False
            return

        self.visit(ast.left)
        self.visit(ast.right)

    def visit_alligned(self, ast: waveform.AlignedWaveform) -> Any:
        self.visit(ast.waveform)

    def visit_negative(self, ast: waveform.Negative) -> Any:
        self.visit(ast.waveform)

    def visit_record(self, ast: waveform.Record) -> Any:
        self.visit(ast.waveform)

    def visit_sample(self, ast: waveform.Sample) -> Any:
        self.visit(ast.waveform)

    def visit_scale(self, ast: waveform.Scale) -> Any:
        self.visit(ast.waveform)

    def visit_slice(self, ast: waveform.Slice) -> Any:
        self.visit(ast.waveform)

    def visit_smooth(self, ast: waveform.Smooth) -> Any:
        self.visit(ast.waveform)

    def emit(self, ast: waveform.Waveform) -> IsConstantWaveformResult:
        self.visit(ast)
        duration = ast.duration(**self.assignments)
        value = ast.eval_decimal(duration, **self.assignments)

        wf = waveform.Constant(value, duration)

        return IsConstantWaveformResult(self.is_constant, wf)


@dataclass(frozen=True)
class IsConstantAnalogCircuitResult:
    is_constant: bool
    effective_analog_circuit: analog_circuit.AnalogCircuit


class IsConstantAnalogCircuit(AnalogCircuitVisitor):
    # Note that this visitor is not complete, it only handles the cases that are
    # relevant for the current IR generated by the Builder semantics.

    @beartype
    def __init__(self, assignments: Dict[str, Decimal] = {}) -> None:
        self.assignments = dict(assignments)
        self.is_constant = True
        self.duration = None

    def visit_waveform(self, ast: waveform.Waveform):
        result = IsConstantWaveform(self.assignments).emit(ast)
        self.is_constant = self.is_constant and result.is_constant
        self.waveform = result.constant_waveform
        if self.duration is None:
            self.duration = self.waveform.duration(**self.assignments)
        else:
            if self.duration != self.waveform.duration(**self.assignments):
                self.is_constant = False

    def visit_field(self, ast: field.Field) -> Any:
        self.field = field.Field({})
        for sm, wf in ast.drives.items():
            self.visit(wf)

            self.field = self.field.add(field.Field({sm: self.waveform}))

    def visit_pulse(self, ast: pulse.Pulse) -> pulse.Pulse:
        self.pulse = pulse.Pulse({})
        for fn, fd in ast.fields.items():
            self.visit(fd)
            self.pulse.fields[fn] = self.field

    def visit_sequence(self, ast: sequence.Sequence) -> sequence.Sequence:
        self.sequence = sequence.Sequence({})
        for pn, ps in ast.pulses.items():
            self.visit(ps)
            self.sequence.pulses[pn] = self.pulse

    def visit_analog_circuit(self, ast: analog_circuit.AnalogCircuit) -> Any:
        self.visit(ast.sequence)
        self.analog_circuit = analog_circuit.AnalogCircuit(ast.register, self.sequence)

    def emit(self, ast: analog_circuit.AnalogCircuit) -> IsConstantAnalogCircuitResult:
        self.visit(ast)
        return IsConstantAnalogCircuitResult(self.is_constant, self.analog_circuit)



from bloqade.builder.typing import ScalarType
from beartype.typing import List, Tuple, Optional, Union, TYPE_CHECKING
from beartype.vale import Is
from typing import Annotated
from beartype import beartype
from plum import dispatch
import numpy as np
from bloqade.ir.scalar import cast

if TYPE_CHECKING:
    from bloqade.ir.location.list import ListOfLocations


def check_position_array(array):
    return (
        array.ndim == 2
        and (
            np.issubdtype(array.dtype, np.floating)
            or np.issubdtype(array.dtype, np.integer)
        )
        and array.shape[1] == 2
    )


def check_bool_array(array):
    return array.ndim == 1 and np.issubdtype(array.dtype, np.bool_)


PositionArray = Annotated[np.ndarray, Is[check_position_array]]
BoolArray = Annotated[np.ndarray, Is[check_bool_array]]


class TransformTrait:
    @beartype
    def scale(self, scale: ScalarType):
        """
        Scale the geometry of your atoms.

        ### Usage Example:
        ```
        >>> reg = start.add_position([(0,0), (1,1)])
        # atom positions are now (0,0), (2,2)
        >>> new_reg = reg.scale(2)
        # you may also use scale on pre-defined geometries
        >>> from bloqade.atom_arrangement import Chain
        # atoms in the chain will now be 2 um apart versus
        # the default 1 um
        >>> Chain(11).scale(2)
        ```

        - Next possible steps are:
        - Continuing to build your geometry via:
            - `...add_position(positions).add_position(positions)`:
                to add more positions
            - `...add_position(positions).apply_defect_count(n_defects)`:
            to randomly drop out n_atoms
            - `...add_position(positions).apply_defect_density(defect_probability)`:
            to drop out atoms with a certain probability
            - `...add_position(positions).scale(scale)`: to scale the geometry
        - Targeting a level coupling once you're done with the atom geometry:
            - `...add_position(positions).rydberg`:
            to specify Rydberg coupling
            - `...add_position(positions).hyperfine`:
            to specify Hyperfine coupling
        - Visualizing your atom geometry:
            - `...add_position(positions).show()`:
            shows your geometry in your web browser

        """
        from .list import ListOfLocations
        from .base import LocationInfo

        scale = cast(scale)
        location_list = []
        for location_info in self.enumerate():
            x, y = location_info.position
            new_position = (scale * x, scale * y)
            location_list.append(
                LocationInfo(new_position, bool(location_info.filling.value))
            )

        return ListOfLocations(location_list)

    @dispatch
    def _add_position(
        self, position: Tuple[ScalarType, ScalarType], filling: Optional[bool] = None
    ):
        from .list import ListOfLocations
        from .base import LocationInfo

        if filling is None:
            filling = True

        location_list = list(self.enumerate())
        location_list.append(LocationInfo(position, filling))

        return ListOfLocations(location_list)

    @dispatch
    def _add_position(  # noqa: F811
        self,
        position: List[Tuple[ScalarType, ScalarType]],
        filling: Optional[List[bool]] = None,
    ):
        from .list import ListOfLocations
        from .base import LocationInfo

        location_list = list(self.enumerate())

        assert (
            len(position) == len(filling) if filling else True
        ), "Length of positions and filling must be the same"

        if filling:
            for position, filling in zip(position, filling):
                location_list.append(LocationInfo(position, filling))

        else:
            for position in position:
                location_list.append(LocationInfo(position, True))

        return ListOfLocations(location_list)

    @dispatch
    def _add_position(  # noqa: F811
        self, position: PositionArray, filling: Optional[BoolArray] = None
    ):
        return self.add_position(
            list(map(tuple, position.tolist())),
            filling.tolist() if filling is not None else None,
        )

    def add_position(
        self,
        position: Union[
            PositionArray,
            List[Tuple[ScalarType, ScalarType]],
            Tuple[ScalarType, ScalarType],
        ],
        filling: Optional[Union[BoolArray, List[bool], bool]] = None,
    ) -> "ListOfLocations":
        """
        Add a position or multiple positions to a pre-existing geometry.

        `add_position` is capable of accepting:
        - A single tuple for one atom coordinate: `(1.0, 2.5)`
        - A list of tuples: `[(0.0, 1.0), (2.0,1.5), etc.]
        - A numpy array of shape (N, 2) where N is the number of atoms

        You may also intersperse variables anywhere a value may be present.

        You can also pass in an optional argument which determines the atom "filling"
        (whether or not at a specified coordinate an atom should be present).

        ### Usage Example:
        ```
        # single coordinate
        >>> reg = start.add_position((0,0))
        # you may chain add_position calls
        >>> reg_plus_two = reg.add_position([(2,2),(5.0, 2.1)])
        # you can add variables anywhere a value may be present
        >>> reg_with_var = reg_plus_two.add_position(("x", "y"))
        # and specify your atom fillings
        >>> reg_with_filling = reg_with_var.add_position([(3.1, 0.0), (4.1, 2.2)],
        [True, False])
        # alternatively you could use one boolean to specify
        # all coordinates should be empty/filled
        >>> reg_with_more_filling = reg_with_filling.add_positions([(3.1, 2.9),
        (5.2, 2.2)], False)
        ```

        - Next possible steps are:
        - Continuing to build your geometry via:
            - `...add_position(positions).add_position(positions)`:
                to add more positions
            - `...add_position(positions).apply_defect_count(n_defects)`:
            to randomly drop out n_atoms
            - `...add_position(positions).apply_defect_density(defect_probability)`:
            to drop out atoms with a certain probability
            - `...add_position(positions).scale(scale)`: to scale the geometry
        - Targeting a level coupling once you're done with the atom geometry:
            - `...add_position(positions).rydberg`: to specify Rydberg coupling
            - `...add_position(positions).hyperfine`: to specify Hyperfine coupling
        - Visualizing your atom geometry:
            - `...add_position(positions).show()`:
            shows your geometry in your web browser

        """
        return self._add_position(position, filling)

    @beartype
    def apply_defect_count(
        self, n_defects: int, rng: np.random.Generator = np.random.default_rng()
    ):
        """
        Drop `n_defects` atoms from the geometry randomly. Internally this occurs
        by setting certain sites to have a SiteFilling set to false indicating
        no atom is present at the coordinate.

        A default numpy-based Random Number Generator is used but you can
        explicitly override this by passing in your own.

        ### Usage Example:

        ```
        >>> from bloqade.atom_arrangement import Chain
        >>> import numpy as np
        # set a custom seed for a numpy-based RNG
        >>> custom_rng = np.random.default_rng(888)
        # randomly remove two atoms from the geometry
        >>> reg = Chain(11).apply_defect_count(2, custom_rng)
        # you may also chain apply_defect_count calls
        >>> reg.apply_defect_count(2, custom_rng)
        # you can also use apply_defect_count on custom geometries
        >>> from bloqade import start
        >>> start.add_position([(0,0), (1,1)]).apply_defect_count(1, custom_rng)
        ```

        - Next possible steps are:
        - Continuing to build your geometry via:
            - `...apply_defect_count(defect_counts).add_position(positions)`:
                to add more positions
            - `...apply_defect_count(defect_counts)
                .apply_defect_count(n_defects)`: to randomly drop out n_atoms
            - `...apply_defect_count(defect_counts)
                .apply_defect_density(defect_probability)`:
                to drop out atoms with a certain probability
            - `...apply_defect_count(defect_counts).scale(scale)`:
                to scale the geometry
        - Targeting a level coupling once you're done with the atom geometry:
            - `...apply_defect_count(defect_counts).rydberg`: to specify
                Rydberg coupling
            - `...apply_defect_count(defect_counts).hyperfine`:
                to specify Hyperfine coupling
        - Visualizing your atom geometry:
            - `...apply_defect_count(defect_counts).show()`:
                shows your geometry in your web browser
        """
        from .list import ListOfLocations
        from .base import LocationInfo, SiteFilling

        location_list = []
        for location_info in self.enumerate():
            location_list.append(location_info)

        filled_sites = []

        for index, location_info in enumerate(location_list):
            if location_info.filling is SiteFilling.filled:
                filled_sites.append(index)

        if n_defects >= len(filled_sites):
            raise ValueError(
                f"n_defects {n_defects} must be less than the number of filled sites "
                f"({len(filled_sites)})"
            )

        for _ in range(n_defects):
            index = rng.choice(filled_sites)
            location_list[index] = LocationInfo(
                location_list[index].position,
                (False if location_list[index].filling is SiteFilling.filled else True),
            )
            filled_sites.remove(index)

        return ListOfLocations(location_list)

    @beartype
    def apply_defect_density(
        self,
        defect_probability: float,
        rng: np.random.Generator = np.random.default_rng(),
    ):
        """
        Drop atoms randomly with `defect_probability` probability (range of 0 to 1).
        Internally this occurs by setting certain sites to have a SiteFilling
        set to false indicating no atom is present at the coordinate.

        A default numpy-based Random Number Generator is used but you can
        explicitly override this by passing in your own.

        ### Usage Example:

        ```
        >>> from bloqade.atom_arrangement import Chain
        >>> import numpy as np
        # set a custom seed for a numpy-based RNG
        >>> custom_rng = np.random.default_rng(888)
        # randomly remove two atoms from the geometry
        >>> reg = Chain(11).apply_defect_density(0.2, custom_rng)
        # you may also chain apply_defect_density calls
        >>> reg.apply_defect_count(0.1, custom_rng)
        # you can also use apply_defect_density on custom geometries
        >>> from bloqade import start
        >>> start.add_position([(0,0), (1,1)])
        .apply_defect_density(0.5, custom_rng)
        ```

        - Next possible steps are:
        - Continuing to build your geometry via:
            - `...apply_defect_count(defect_counts).add_position(positions)`:
            to add more positions
            - `...apply_defect_count(defect_counts).apply_defect_count(n_defects)`:
            to randomly drop out n_atoms
            - `...apply_defect_count(defect_counts)
            .apply_defect_density(defect_probability)`:
            to drop out atoms with a certain probability
            - `...apply_defect_count(defect_counts).scale(scale)`:
            to scale the geometry
        - Targeting a level coupling once you're done with the atom geometry:
            - `...apply_defect_count(defect_counts).rydberg`:
            to specify Rydberg coupling
            - `...apply_defect_count(defect_counts).hyperfine`:
            to specify Hyperfine coupling
        - Visualizing your atom geometry:
            - `...apply_defect_count(defect_counts).show()`:
            shows your geometry in your web browser
        """
        from .list import ListOfLocations
        from .base import LocationInfo, SiteFilling

        p = min(1, max(0, defect_probability))
        location_list = []

        for location_info in self.enumerate():
            if rng.random() < p:
                location_list.append(
                    LocationInfo(
                        location_info.position,
                        (
                            False
                            if location_info.filling is SiteFilling.filled
                            else True
                        ),
                    )
                )
            else:
                location_list.append(location_info)

        return ListOfLocations(location_list=location_list)

    def remove_vacant_sites(self):
        from .base import SiteFilling
        from .list import ListOfLocations

        new_locations = []
        for location_info in self.enumerate():
            if location_info.filling is SiteFilling.filled:
                new_locations.append(location_info)

        return ListOfLocations(new_locations)



from bokeh.plotting import figure
from bokeh.models import ColumnDataSource, Step, HoverTool
from bokeh.layouts import gridplot, row


def get_task_ir_figure(task_ir, **fig_kwargs):
    from bloqade.submission.ir.task_specification import (
        RabiFrequencyAmplitude,
        RabiFrequencyPhase,
        Detuning,
        Lattice,
        QuEraTaskSpecification,
    )

    if isinstance(task_ir, RabiFrequencyAmplitude):
        return get_rabi_amp_figure(task_ir, **fig_kwargs)
    elif isinstance(task_ir, RabiFrequencyPhase):
        return get_rabi_phase_figure(task_ir, **fig_kwargs)
    elif isinstance(task_ir, Detuning):
        return get_detune_figure(task_ir, **fig_kwargs)
    elif isinstance(task_ir, Lattice):
        return get_lattice_figure(task_ir, **fig_kwargs)
    elif isinstance(task_ir, QuEraTaskSpecification):
        return get_quera_task_figure(task_ir, **fig_kwargs)
    else:
        raise NotImplementedError(
            f"not supported task IR for figure, got {type(task_ir)}"
        )


def get_lattice_figure(lattice, **fig_kwargs):
    from bloqade.ir.location import ListOfLocations

    ## use ir.Atom_oarrangement's plotting:
    ## covert unit to m -> um
    sites_um = list(
        map(lambda cord: (float(cord[0]) * 1e6, float(cord[1]) * 1e6), lattice.sites)
    )
    reg = ListOfLocations().add_position(sites_um, lattice.filling)
    fig_reg = reg.figure(fig_kwargs=fig_kwargs)  # ignore the B-rad widget
    return fig_reg


def get_quera_task_figure(task_ir, **fig_kwargs):
    # grab global figures
    rabi_amplitude = (
        task_ir.effective_hamiltonian.rydberg.rabi_frequency_amplitude.figure(
            tools="wheel_zoom,reset, undo, redo, pan"
        )
    )
    rabi_phase = task_ir.effective_hamiltonian.rydberg.rabi_frequency_phase.figure(
        x_range=rabi_amplitude.x_range,
        tools="hover,wheel_zoom,reset, undo, redo, pan",
    )
    global_detuning = task_ir.effective_hamiltonian.rydberg.detuning.global_figure(
        x_range=rabi_amplitude.x_range,
        tools="hover,wheel_zoom,reset, undo, redo, pan",
    )

    # lattice:
    register = task_ir.lattice.figure(x_axis_label="x (um)", y_axis_label="y (um)")

    col_plt = gridplot(
        [[rabi_amplitude], [global_detuning], [rabi_phase]],
        merge_tools=False,
        sizing_mode="stretch_both",
    )
    col_plt.width_policy = "max"

    full_plt = row(col_plt, register, sizing_mode="stretch_both")
    full_plt.width_policy = "max"

    return full_plt


def get_rabi_phase_figure(rabi_phase, **fig_kwargs):
    source = ColumnDataSource(rabi_phase._get_data_source())
    TOOLTIPS = [("(x,y)", "(@times_phase, @values_phase)")]

    line_plt = figure(
        **fig_kwargs,
        tooltips=TOOLTIPS,
        x_axis_label="Time (s)",
        y_axis_label="ϕ(t) (rad)",
    )

    line_plt.y_range.start = min(source.data["values_phase"]) - 5e6
    line_plt.y_range.end = max(source.data["values_phase"]) + 5e6
    line_plt.x_range.start = 0

    steps = Step(
        x="times_phase",
        y="values_phase",
        line_color="black",
        line_width=2,
        mode="center",
    )

    line_plt.add_glyph(source, steps)

    line_plt.circle(
        x="times_phase", y="values_phase", source=source, size=4, color="black"
    )

    return line_plt


def get_rabi_amp_figure(rabi_amp, **fig_kwargs):
    source = ColumnDataSource(rabi_amp._get_data_source())
    hover = HoverTool()
    hover.tooltips = [("(x,y)", "(@times_amp, @values_amp)")]

    line_plt = figure(
        **fig_kwargs,
        x_axis_label="Time (s)",
        y_axis_label="Ω(t) (rad/s)",
    )

    line_plt.x_range.start = 0
    line_plt.y_range.start = min(source.data["values_amp"]) - 5e6
    line_plt.y_range.end = max(source.data["values_amp"]) + 5e6

    line_plt.line(
        x="times_amp", y="values_amp", source=source, line_width=2, color="black"
    )

    line_plt.circle(x="times_amp", y="values_amp", source=source, size=4, color="black")

    line_plt.varea(
        x="times_amp",
        y1="values_amp",
        y2="values_floor_amp",
        source=source,
        fill_alpha=0.3,
        color="#6437FF",
    )
    line_plt.add_tools(hover)

    return line_plt


def get_detune_figure(detune, **fig_kwargs):
    source = ColumnDataSource(detune._get_data_source())
    TOOLTIPS = [("(x,y)", "(@times_detune, @values_detune)")]

    line_plt = figure(
        **fig_kwargs,
        tooltips=TOOLTIPS,
        x_axis_label="Time (s)",
        y_axis_label="Δ(t) (rad/s)",
    )

    line_plt.x_range.start = 0

    line_plt.y_range.start = min(source.data["values_detune"]) - 5e6
    line_plt.y_range.end = max(source.data["values_detune"]) + 5e6

    line_plt.line(
        x="times_detune",
        y="values_detune",
        source=source,
        line_width=2,
        color="black",
    )

    line_plt.circle(
        x="times_detune", y="values_detune", source=source, size=4, color="black"
    )

    line_plt.varea(
        x="times_detune",
        y1="values_detune",
        y2="values_floor_detune",
        source=source,
        fill_alpha=0.5,
        color="#EFD0DE",
    )

    return line_plt



from numbers import Real
from bloqade.builder.typing import ScalarType
from bloqade.ir.tree_print import Printer
from bloqade.ir.scalar import (
    Scalar,
    Interval,
    Variable,
    AssignedVariable,
    cast,
    var,
)

from bisect import bisect_left
from decimal import Decimal
from pydantic.dataclasses import dataclass
from beartype.typing import Any, Tuple, Union, List, Callable, Dict
from beartype import beartype
from enum import Enum

import numpy as np
import inspect
import scipy.integrate as integrate
from bloqade.visualization import get_ir_figure
from bloqade.visualization import display_ir
from functools import cached_property


@beartype
def to_waveform(duration: ScalarType) -> Callable[[Callable], "PythonFn"]:
    # turn python function into a waveform instruction."""

    def waveform_wrapper(fn: Callable) -> "PythonFn":
        return PythonFn.create(fn, duration)

    return waveform_wrapper


class AlignedValue(str, Enum):
    Left = "left_value"
    Right = "right_value"


class Alignment(str, Enum):
    Left = "left_aligned"
    Right = "right_aligned"


@dataclass(frozen=True)
class Waveform:
    """
    Waveform node in the IR.

    - [`<instruction>`][bloqade.ir.control.waveform.Instruction]
    - [`<smooth>`][bloqade.ir.control.waveform.Smooth]
    - [`<slice>`][bloqade.ir.control.waveform.Slice]
    - [`<apppend>`][bloqade.ir.control.waveform.Append]
    - [`<negative>`][bloqade.ir.control.waveform.Negative]
    - [`<scale>`][bloqade.ir.control.waveform.Scale]
    - [`<add>`][bloqade.ir.control.waveform.Add]
    - [`<record>`][bloqade.ir.control.waveform.Record]
    - [`<sample>`][bloqade.ir.control.waveform.Sample]

    ```bnf
    <waveform> ::= <instruction>
        | <smooth>
        | <slice>
        | <append>
        | <negative>
        | <scale>
        | <add>
        | <record>
        | <sample>
    ```
    """

    def __call__(self, clock_s: float, **kwargs) -> float:
        return float(self.eval_decimal(Decimal(str(clock_s)), **kwargs))

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        raise NotImplementedError

    def add(self, other: "Waveform") -> "Waveform":
        return self.canonicalize(Add(self, other))

    def append(self, other: "Waveform") -> "Waveform":
        return self.canonicalize(Append([self, other]))

    def figure(self, **assignments):
        """get figure of the plotting the waveform.

        Returns:
            figure: a bokeh figure
        """
        return get_ir_figure(self, **assignments)

    def _get_data(self, npoints, **assignments):
        from bloqade.ir.analysis.assignment_scan import AssignmentScanRecord

        assignments = AssignmentScanRecord(assignments).emit(self)

        duration = float(self.duration(**assignments))
        times = np.linspace(0, duration, npoints + 1)
        values = [self.__call__(time, **assignments) for time in times]
        return times, values

    def show(self, **assignments):
        display_ir(self, assignments)

    def align(
        self, alignment: Alignment, value: Union[None, AlignedValue, Scalar] = None
    ) -> "Waveform":
        if value is None:
            if alignment == Alignment.Left:
                value = AlignedValue.Left
            elif alignment == Alignment.Right:
                value = AlignedValue.Right
            else:
                raise ValueError(f"Invalid alignment: {alignment}")
        elif value is Scalar or value is AlignedValue:
            return self.canonicalize(AlignedWaveform(self, alignment, value))
        else:
            return self.canonicalize(AlignedWaveform(self, alignment, cast(value)))

    def smooth(self, radius, kernel: "SmoothingKernel") -> "Waveform":
        return self.canonicalize(
            Smooth(kernel=kernel, waveform=self, radius=cast(radius))
        )

    def scale(self, value) -> "Waveform":
        return self.canonicalize(Scale(cast(value), self))

    def __neg__(self) -> "Waveform":
        return self.canonicalize(Negative(self))

    def __getitem__(self, s: slice) -> "Waveform":
        return self.canonicalize(Slice(self, Interval.from_slice(s)))

    def record(self, variable_name: Union[str, Variable]):
        return Record(self, cast(variable_name))

    def __add__(self, other: "Waveform") -> "Waveform":
        if isinstance(other, Waveform):
            return self.canonicalize(Add(self, other))

        return NotImplemented

    def __radd__(self, other: "Waveform") -> "Waveform":
        if isinstance(other, Waveform):
            return self.canonicalize(Add(self, other))

        return NotImplemented

    def __sub__(self, other: "Waveform") -> "Waveform":
        if isinstance(other, Waveform):
            return self + (-other)

        return NotImplemented

    def __rsubs__(self, other: "Waveform") -> "Waveform":
        if isinstance(other, Waveform):
            return other + (-self)

        return NotImplemented

    def __mul__(self, other: Any) -> "Waveform":
        return self.scale(cast(other))

    def __rmul__(self, other: Any) -> "Waveform":
        return self.scale(cast(other))

    def __truediv__(self, other: Any) -> "Waveform":
        return self.scale(1 / cast(other))

    def __str__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    @staticmethod
    def canonicalize(expr: "Waveform") -> "Waveform":
        if isinstance(expr, Append):
            new_waveforms = []
            for waveform in expr.waveforms:
                if isinstance(waveform, Append):
                    new_waveforms += waveform.waveforms
                else:
                    new_waveforms.append(waveform)

            new_waveforms = list(map(Waveform.canonicalize, new_waveforms))
            return Append(new_waveforms)
        elif isinstance(expr, Add):
            left = Waveform.canonicalize(expr.left)
            right = Waveform.canonicalize(expr.right)
            if left == right:
                return left.scale(2)
            if left.duration == cast(0):
                return right
            if right.duration == cast(0):
                return left
            else:
                return expr
        else:
            return expr

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def print_node(self):
        raise NotImplementedError


@dataclass(frozen=True)
class AlignedWaveform(Waveform):
    """

    ```bnf
    <padded waveform> ::= <waveform> | <waveform> <alignment> <value>

    <alignment> ::= 'left aligned' | 'right aligned'
    <value> ::= 'left value' | 'right value' | <scalar expr>
    ```
    """

    waveform: Waveform
    alignment: Alignment
    value: Union[Scalar, AlignedValue]

    @property
    def duration(self):
        return self.waveform.duration

    def print_node(self):
        return "AlignedWaveform"

    def children(self):
        annotated_children = {}
        annotated_children["Waveform"] = self.waveform

        if self.alignment == Alignment.Left:
            annotated_children["Alignment"] = "Left"
        elif self.alignment == Alignment.Right:
            annotated_children["Alignment"] = "Right"

        if isinstance(self.value, Scalar):
            annotated_children["Value"] = self.value
        elif self.value == AlignedValue.Left:
            annotated_children["Value"] = "Left"
        elif self.value == AlignedValue.Right:
            annotated_children["Value"] = "Right"

        return annotated_children


@dataclass(frozen=True)
class Instruction(Waveform):
    """Instruction node in the IR.

    - [`<linear>`][bloqade.ir.control.waveform.Linear]
    - [`<constant>`][bloqade.ir.control.waveform.Constant]
    - [`<poly>`][bloqade.ir.control.waveform.Poly]
    - [`<python-fn>`][bloqade.ir.control.waveform.PythonFn]


    ```bnf
    <instruction> ::= <linear>
        | <constant>
        | <poly>
        | <python-fn>
    ```
    """

    pass


@dataclass(init=False, frozen=True)
class Linear(Instruction):
    """
    ```bnf
    <linear> ::= 'linear' <scalar expr> <scalar expr>
    ```

    f(t=0:duration) = start + (stop-start)/duration * t

    Args:
        start (Scalar): start value
        stop (Scalar): stop value
        duration (Scalar): the time span of the linear waveform.

    """

    start: Scalar
    stop: Scalar
    duration: Scalar

    @beartype
    def __init__(self, start: ScalarType, stop: ScalarType, duration: ScalarType):
        object.__setattr__(self, "start", cast(start))
        object.__setattr__(self, "stop", cast(stop))
        object.__setattr__(self, "duration", cast(duration))

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        start_value = self.start(**kwargs)
        stop_value = self.stop(**kwargs)

        if clock_s > self.duration(**kwargs):
            return Decimal(0)
        else:
            return (
                (stop_value - start_value) / self.duration(**kwargs)
            ) * clock_s + start_value

    def print_node(self):
        return "Linear"

    def children(self):
        return {"start": self.start, "stop": self.stop, "duration": self.duration}


@dataclass(init=False, frozen=True)
class Constant(Instruction):
    """
    ```bnf
    <constant> ::= 'constant' <scalar expr>
    ```

    f(t=0:duration) = value

    Args:
        value (Scalar): the constant value
        duration (Scalar): the time span of the constant waveform.

    """

    value: Scalar
    duration: Scalar

    @beartype
    def __init__(self, value: ScalarType, duration: ScalarType):
        object.__setattr__(self, "value", cast(value))
        object.__setattr__(self, "duration", cast(duration))

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        constant_value = self.value(**kwargs)
        if clock_s > self.duration(**kwargs):
            return Decimal(0)
        else:
            return constant_value

    def print_node(self):
        return "Constant"

    def children(self):
        return {"value": self.value, "duration": self.duration}


@dataclass(init=False, frozen=True)
class Poly(Instruction):
    """
    ```bnf
    <poly> ::= <scalar>+
    ```

    f(t=0:duration) = c[0] + c[1]t + c[2]t^2 + ... + c[n-1]t^n-1 + c[n]t^n

    Args:
        coeffs (List[Scalar]): the coefficients c[] of the polynomial.
        duration (Scalar): the time span of the waveform.

    """

    coeffs: Tuple[Scalar, ...]
    duration: Scalar

    @beartype
    def __init__(self, coeffs: List[ScalarType], duration: ScalarType):
        object.__setattr__(self, "coeffs", tuple(map(cast, coeffs)))
        object.__setattr__(self, "duration", cast(duration))

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        # b + x + x^2 + ... + x^n-1 + x^n
        if clock_s > self.duration(**kwargs):
            return Decimal(0)
        else:
            # call clock_s on each element of the scalars,
            # then apply the proper powers
            value = Decimal(0)
            power = Decimal(1)
            for scalar_expr in self.coeffs:
                value += scalar_expr(**kwargs) * power
                power *= clock_s

            return value

    def print_node(self) -> str:
        return "Poly"

    def children(self):
        # should have annotation for duration
        # then annotations for the polynomial terms and exponents

        annotated_coeffs = {}
        for i, coeff in enumerate(self.coeffs):
            if i == 0:
                annotated_coeffs["b"] = coeff
            elif i == 1:
                annotated_coeffs["t"] = coeff
            else:
                annotated_coeffs["t^" + str(i)] = coeff

        annotated_coeffs["duration"] = self.duration
        return annotated_coeffs


@dataclass(frozen=True)
class PythonFn(Instruction):
    """

    ```bnf
    <python-fn> ::= 'python-fn' <python function def> <scalar expr>
    ```
    """

    fn: Callable  # [[float, ...], float] # f(t) -> value
    duration: Scalar
    parameters: List[Union[Variable, AssignedVariable]]  # come from ast inspect
    default_param_values: Dict[str, Decimal]

    @staticmethod
    def create(fn: Callable, duration: ScalarType) -> "PythonFn":
        signature = inspect.getfullargspec(fn)

        if signature.varargs is not None:
            raise ValueError("Cannot have `*args` in function definition")

        if signature.varkw is not None:
            raise ValueError("Cannot have `**kwargs` in function definition")

        # get default kwonly first:
        variables = []
        default_param_values = {}
        if signature.kwonlydefaults is not None:
            for name, value in signature.kwonlydefaults.items():
                if isinstance(value, (Real, Decimal)):
                    variables.append(name)
                    default_param_values[name] = Decimal(str(value))
                else:
                    raise ValueError(
                        f"Default value for parameter {name} is not Real or Decimal, "
                        "cannot convert to Variable."
                    )

        variables += signature.args[1:]
        variables += signature.kwonlyargs

        parameters = list(map(var, variables))
        duration = cast(duration)

        return PythonFn(fn, duration, parameters, default_param_values)

    @cached_property
    def _hash_value(self) -> int:
        return (
            hash(self.__class__)
            ^ hash(self.fn)
            ^ hash(self.duration)
            ^ hash(tuple(self.parameters))
            ^ hash(frozenset(self.default_param_values.items()))
        )

    def __hash__(self) -> int:
        return self._hash_value

    def eval_decimal(self, clock_s: Decimal, **assignments) -> Decimal:
        new_assignments = {**self.default_param_values, **assignments}

        if clock_s > self.duration(**new_assignments):
            return Decimal(0)

        kwargs = {
            param.name: float(param(**new_assignments)) for param in self.parameters
        }
        return Decimal(
            str(
                self.fn(
                    float(clock_s),
                    **kwargs,
                )
            )
        )

    def print_node(self):
        return f"PythonFn: {self.fn.__name__}"

    def children(self):
        return {"duration": self.duration, **{p.name: p for p in self.parameters}}

    def sample(
        self, dt: ScalarType, interpolation: Union[str, "Interpolation"]
    ) -> "Sample":
        return Sample(self, interpolation, cast(dt))


@dataclass(frozen=True)
class SmoothingKernel:
    def __call__(self, value: float) -> float:
        raise NotImplementedError


class FiniteSmoothingKernel(SmoothingKernel):
    # kernel that is zero outside of (-1, 1)
    pass


class InfiniteSmoothingKernel(SmoothingKernel):
    # Kernel that is non-zero for all values
    pass


@dataclass(frozen=True)
class Gaussian(InfiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return np.exp(-(value**2) / 2) / np.sqrt(2 * np.pi)


@dataclass(frozen=True)
class Logistic(InfiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return np.exp(-(np.logaddexp(0, value) + np.logaddexp(0, -value)))


@dataclass(frozen=True)
class Sigmoid(InfiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return (2 / np.pi) * np.exp(-np.logaddexp(-value, value))


@dataclass(frozen=True)
class Triangle(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return np.maximum(0, 1 - np.abs(value))


@dataclass(frozen=True)
class Uniform(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return np.asarray(np.abs(value) <= 1, dtype=np.float64).squeeze()


@dataclass(frozen=True)
class Parabolic(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return (3 / 4) * np.maximum(0, 1 - value**2)


@dataclass(frozen=True)
class Biweight(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return (15 / 16) * np.maximum(0, 1 - value**2) ** 2


@dataclass(frozen=True)
class Triweight(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return (35 / 32) * np.maximum(0, 1 - value**2) ** 3


@dataclass(frozen=True)
class Tricube(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return (70 / 81) * np.maximum(0, 1 - np.abs(value) ** 3) ** 3


@dataclass(frozen=True)
class Cosine(FiniteSmoothingKernel):
    def __call__(self, value: float) -> float:
        return np.maximum(0, np.pi / 4 * np.cos(np.pi / 2 * value))


GaussianKernel = Gaussian()
LogisticKernel = Logistic()
SigmoidKernel = Sigmoid()
TriangleKernel = Triangle()
UniformKernel = Uniform()
ParabolicKernel = Parabolic()
BiweightKernel = Biweight()
TriweightKernel = Triweight()
TricubeKernel = Tricube()
CosineKernel = Cosine()


@dataclass(init=False, frozen=True)
class Smooth(Waveform):
    """
    ```bnf
    <smooth> ::= 'smooth' <kernel> <waveform>
    ```
    """

    radius: Scalar
    kernel: SmoothingKernel
    waveform: Waveform

    def __init__(self, radius, kernel, waveform):
        if isinstance(kernel, str):
            if kernel == "Guassian":
                kernel = GaussianKernel
            elif kernel == "Logistic":
                kernel = LogisticKernel
            elif kernel == "Sigmoid":
                kernel = SigmoidKernel
            elif kernel == "Triangle":
                kernel = TriangleKernel
            elif kernel == "Uniform":
                kernel = UniformKernel
            elif kernel == "Parabolic":
                kernel = ParabolicKernel
            elif kernel == "Biweight":
                kernel = BiweightKernel
            elif kernel == "Triweight":
                kernel = TriweightKernel
            elif kernel == "Tricube":
                kernel = TricubeKernel
            elif kernel == "Cosine":
                kernel = CosineKernel
            else:
                raise ValueError(f"Invalid kernel: {kernel}")

        object.__setattr__(self, "radius", cast(radius))
        object.__setattr__(self, "kernel", kernel)
        object.__setattr__(self, "waveform", waveform)

    @property
    def duration(self):
        return self.waveform.duration

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        float_clock_s = float(clock_s)
        radius = float(self.radius(**kwargs))
        duration = float(self.duration(**kwargs))
        waveform_start = self.waveform(0, **kwargs)
        waveform_stop = self.waveform(float(duration), **kwargs)

        def waveform(clock: float) -> float:
            if clock < 0:
                return waveform_start
            elif clock > duration:
                return waveform_stop
            else:
                return self.waveform(clock, **kwargs)

        def integrade(s):
            return self.kernel(s) * waveform(radius * s + float_clock_s)

        if isinstance(self.kernel, FiniteSmoothingKernel):
            return integrate.quad(integrade, -1, 1, epsabs=1e-4, epsrel=1e-4)[0]
        elif isinstance(self.kernel, InfiniteSmoothingKernel):
            return integrate.quad(integrade, -np.inf, np.inf, epsabs=1e-4, epsrel=1e-4)[
                0
            ]
        else:
            raise ValueError(f"Invalid kernel: {self.kernel}")


@dataclass(frozen=True)
class Slice(Waveform):
    """
    ```
    <slice> ::= <waveform> <scalar.interval>
    ```
    """

    waveform: Waveform
    interval: Interval

    @cached_property
    def duration(self):
        from bloqade.ir.scalar import Slice

        if self.interval.start is None and self.interval.stop is None:
            raise ValueError("Interval must have a start or stop value")

        return Slice(self.waveform.duration, self.interval)

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        if clock_s > self.duration(**kwargs):
            return Decimal(0)
        if self.interval.start is None:
            start_time = Decimal(0)
        else:
            start_time = self.interval.start(**kwargs)
        return self.waveform.eval_decimal(clock_s + start_time, **kwargs)

    def print_node(self):
        return "Slice"

    def children(self):
        return [self.waveform, self.interval]


@dataclass(frozen=True)
class Append(Waveform):
    """
    ```bnf
    <append> ::= <waveform>+
    ```
    """

    waveforms: Tuple[Waveform, ...]

    @cached_property
    def duration(self):
        duration = cast(0.0)
        for waveform in self.waveforms:
            duration = duration + waveform.duration

        return duration

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        append_time = Decimal(0)
        for waveform in self.waveforms:
            duration = waveform.duration(**kwargs)

            if clock_s <= append_time + duration:
                return waveform.eval_decimal(clock_s - append_time, **kwargs)

            append_time += duration

        return Decimal(0)

    def print_node(self):
        return "Append"

    def children(self):
        return self.waveforms


@dataclass(frozen=True)
class Negative(Waveform):
    """
    ```bnf
    <negative> ::= '-' <waveform>
    ```
    """

    waveform: Waveform

    @property
    def duration(self):
        return self.waveform.duration

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        return -self.waveform.eval_decimal(clock_s, **kwargs)

    def print_node(self):
        return "Negative"

    def children(self):
        return [self.waveform]


@dataclass(init=False, frozen=True)
class Scale(Waveform):
    """
    ```bnf
    <scale> ::= <scalar expr> '*' <waveform>
    ```
    """

    scalar: Scalar
    waveform: Waveform

    def __init__(self, scalar, waveform: Waveform):
        object.__setattr__(self, "scalar", cast(scalar))
        object.__setattr__(self, "waveform", waveform)

    @property
    def duration(self):
        return self.waveform.duration

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        return self.scalar(**kwargs) * self.waveform.eval_decimal(clock_s, **kwargs)

    def print_node(self):
        return "Scale"

    def children(self):
        return [self.scalar, self.waveform]


@dataclass(frozen=True)
class Add(Waveform):
    """
    ```bnf
    <add> ::= <waveform> '+' <waveform>
    ```
    """

    left: Waveform
    right: Waveform

    @cached_property
    def duration(self):
        return self.left.duration.max(self.right.duration)

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        return self.left(clock_s, **kwargs) + self.right(clock_s, **kwargs)

    def print_node(self):
        return "+"

    def children(self):
        return [self.left, self.right]


@dataclass(frozen=True)
class Record(Waveform):
    """
    ```bnf
    <record> ::= 'record' <waveform> <var>
    ```
    """

    waveform: Waveform
    var: Variable

    @property
    def duration(self):
        return self.waveform.duration

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        return self.waveform(clock_s, **kwargs)

    def print_node(self):
        return "Record"

    def children(self):
        return {"Waveform": self.waveform, "Variable": self.var}


class Interpolation(str, Enum):
    Linear = "linear"
    Constant = "constant"


@dataclass(frozen=True)
class Sample(Waveform):
    """
    ```bnf
    <sample> ::= 'sample' <waveform> <interpolation> <scalar>
    ```
    """

    waveform: Waveform
    interpolation: Interpolation
    dt: Scalar

    @property
    def duration(self):
        return self.waveform.duration

    def samples(self, **kwargs) -> Tuple[List[Decimal], List[Decimal]]:
        duration = self.duration(**kwargs)
        dt = self.dt(**kwargs)

        clock = Decimal("0.0")
        clocks = []
        values = []
        while clock <= duration - dt:
            values.append(self.waveform.eval_decimal(clock, **kwargs))
            clocks.append(clock)
            clock += dt

        values.append(self.waveform.eval_decimal(duration, **kwargs))
        clocks.append(duration)

        return clocks, values

    def eval_decimal(self, clock_s: Decimal, **kwargs) -> Decimal:
        times, values = self.samples(**kwargs)
        i = bisect_left(times, clock_s)

        if i == len(times):
            return Decimal(0)

        if self.interpolation is Interpolation.Linear:
            if i == 0:
                return values[i]
            else:
                slope = (values[i] - values[i - 1]) / (times[i] - times[i - 1])
                return slope * (clock_s - times[i - 1]) + values[i - 1]

        elif self.interpolation is Interpolation.Constant:
            if i == 0:
                return values[i]
            else:
                return values[i - 1]

    def print_node(self):
        return f"Sample {self.interpolation.value}"

    def children(self):
        return {"Waveform": self.waveform, "sample_step": self.dt}



from bloqade.serialize import dumps, loads
from bloqade import start
from bloqade.task.quera import QuEraTask
from bloqade.task.braket import BraketTask
from bloqade.submission.quera import QuEraBackend
from bloqade.submission.braket import BraketBackend
import pytest


def test_quera_task():
    backend = QuEraBackend(api_hostname="a", qpu_id="b")
    task = QuEraTask(
        task_ir=None, task_id=None, metadata={}, backend=backend, parallel_decoder=None
    )

    with pytest.raises(ValueError):
        task.fetch()

    assert task._result_exists() is False

    with pytest.raises(ValueError):
        task.pull()


def test_braket_task():
    backend = BraketBackend()
    task = BraketTask(
        task_ir=None, task_id=None, metadata={}, backend=backend, parallel_decoder=None
    )

    with pytest.raises(ValueError):
        task.fetch()

    assert task._result_exists() is False

    with pytest.raises(ValueError):
        task.pull()


def test_braket_batch():
    program = (
        start.add_position((0, 0))
        .rydberg.rabi.amplitude.uniform.piecewise_linear(
            durations=[0.05, 1, 0.05], values=[0.0, 15.8, 15.8, 0.0]
        )
        .detuning.uniform.piecewise_linear(durations=[1.1], values=[0.0, 0.0])
    )

    output = program.braket.aquila()._compile(10)

    output_str = dumps(output)
    assert isinstance(loads(output_str), type(output))
    assert loads(output_str).tasks == output.tasks
    assert loads(output_str).name == output.name


#
# batch = RemoteBatch.from_json(...)
# batch.fetch() # update results,
#               # this is non-blocking.
#               # It only pull results if the remote job complete

# batch.pull() # this is blocking. it will hanging there waiting for job to complete.
# batch.to_json(...)

# # Get finished tasks (not nessasry sucessfully completed)
# finished_batch = batch.get_finished_tasks()

# # Get complete tasks (sucessfully completed)
# completed_batch = batch.get_completed_tasks()

# # Remove failed tasks:
# new_batch = batch.remove_failed_tasks()



from pydantic import ValidationError
from bloqade import cast, var
import bloqade.ir.scalar as scalar
import pytest
from decimal import Decimal
from io import StringIO
from IPython.lib.pretty import PrettyPrinter as PP
import bloqade.ir.tree_print as trp

trp.color_enabled = False


def test_var():
    assert var("a") == scalar.Variable("a")
    assert var("a") != scalar.Variable("b")

    with pytest.raises(TypeError):
        var(1)

    with pytest.raises(ValueError):
        var("a*b")

    with pytest.raises(ValidationError):
        var("__batch_params")

    Vv = var("a")
    vs = var(Vv)

    assert vs == Vv

    mystdout = StringIO()
    p = PP(mystdout)
    Vv._repr_pretty_(p, 0)

    assert mystdout.getvalue() == "Variable: a\n"


def test_cast():
    assert cast("a") == scalar.Variable("a")
    assert cast(1) == scalar.Literal(Decimal("1"))
    assert cast(1.20391023) == scalar.Literal(Decimal("1.20391023"))

    with pytest.raises(ValueError):
        cast("a-b")

    mystdout = StringIO()
    p = PP(mystdout)
    cast(1)._repr_pretty_(p, 0)

    assert mystdout.getvalue() == "Literal: 1\n"


def test_add():
    assert var("a") + var("b") == scalar.Add(var("a"), var("b"))
    assert var("a") + 1 == scalar.Add(var("a"), scalar.Literal(Decimal("1")))
    assert 0 + var("a") == var("a")
    assert var("a") + 0 == var("a")
    assert 1 + var("a") == scalar.Add(scalar.Literal(Decimal("1")), var("a"))
    assert cast(1) + cast(2) == scalar.Literal(Decimal("3"))

    mystdout = StringIO()
    p = PP(mystdout)
    (1 + var("a"))._repr_pretty_(p, 0)

    assert mystdout.getvalue() == "+\n├─ Literal: 1\n⋮\n└─ Variable: a⋮\n"


def test_sub():
    assert var("a") - var("b") == scalar.Add(var("a"), scalar.Negative(var("b")))
    assert var("a") - 1 == scalar.Add(var("a"), scalar.Negative(cast(1)))
    assert 1 - var("a") == scalar.Add(
        scalar.Literal(Decimal("1")), scalar.Negative(var("a"))
    )
    assert cast(1) - cast(2) == scalar.Literal(Decimal("-1"))
    assert -cast(1) + cast(2) == scalar.Literal(Decimal("1"))


def test_mul():
    assert var("a") * var("b") == scalar.Mul(var("a"), var("b"))
    assert 1 * var("a") == var("a")
    assert var("a") * 1 == var("a")
    assert 3 * var("a") == scalar.Mul(scalar.Literal(Decimal("3")), var("a"))
    assert cast(1) * cast(2) == scalar.Literal(Decimal("2"))

    mystdout = StringIO()
    p = PP(mystdout)
    (3 * var("a"))._repr_pretty_(p, 0)

    assert mystdout.getvalue() == "*\n├─ Literal: 3\n⋮\n└─ Variable: a⋮\n"


def test_div():
    assert var("a") / var("b") == scalar.Div(var("a"), var("b"))
    assert 1 / var("a") == scalar.Div(cast(1), var("a"))
    assert var("a") / 1 == var("a")
    assert 3 / var("a") == scalar.Div(cast(3), var("a"))
    assert cast(1) / cast(2) == cast(0.5)

    mystdout = StringIO()
    p = PP(mystdout)
    (3 / var("a"))._repr_pretty_(p, 0)

    assert mystdout.getvalue() == "/\n├─ Literal: 3\n⋮\n└─ Variable: a⋮\n"


@pytest.mark.skip(reason="no longer supported")
def test_list_of_var():
    pylist = ["a", "b", "c"]
    vlist = var(pylist)
    for pyobj, bobj in zip(pylist, vlist):
        assert bobj == var(pyobj)


@pytest.mark.skip(reason="no longer supported")
def test_tuple_of_var():
    pylist = ("a", "b", "c")
    vlist = var(pylist)
    for pyobj, bobj in zip(pylist, vlist):
        assert bobj == var(pyobj)


def test_var_member():
    va = var("a")

    assert va.children() == []
    assert va.print_node() == "Variable: a"
    assert str(va) == "a"


def test_assign_var_member():
    va = scalar.AssignedVariable("a", Decimal("1.0"))

    assert va.children() == []
    assert va.print_node() == "AssignedVariable: a = 1.0"
    assert str(va) == "a"


def test_literal_value():
    va = cast(1.0)
    with pytest.raises(TypeError):
        var(va)

    assert va.children() == []
    assert va.print_node() == "Literal: 1.0"
    assert str(va) == "1.0"


def test_invalid_keyword():
    with pytest.raises(ValidationError):
        var("__batch_params")


def test_negative_node():
    sa = cast(1.0)
    nsa = scalar.Negative(sa)

    assert nsa.children() == [sa]
    assert nsa.print_node() == "Negative"
    assert str(nsa) == "-(1.0)"


# def test_base_invalid():


def test_interval():
    itvl = scalar.Interval(start=cast(0.0), stop=cast(1.0))
    itvl_no_start = scalar.Interval(None, stop=cast(1.0))
    itvl_no_stop = scalar.Interval(start=cast(0), stop=None)
    itvl_wrong = scalar.Interval(None, None)

    assert itvl.print_node() == "Interval"

    with pytest.raises(ValueError):
        str(itvl_wrong)

    with pytest.raises(ValueError):
        itvl_wrong.children()

    assert itvl.children() == {"start": cast(0.0), "stop": cast(1.0)}
    assert itvl_no_start.children() == {"stop": cast(1.0)}
    assert itvl_no_stop.children() == {"start": cast(0)}
    assert str(itvl) == "0.0:1.0"
    assert str(itvl_no_start) == ":1.0"
    assert str(itvl_no_stop) == "0:"


def test_canonicalize_neg_neg():
    sa = cast(1.0)
    nsa = -sa
    nsa2 = -nsa

    out2 = scalar.Scalar.canonicalize(nsa2)
    assert out2.value == Decimal("1.0")

    sb = cast(-1.0)
    nsb = scalar.Negative(sb)

    out3 = scalar.Scalar.canonicalize(nsb)
    assert out3 == cast(1.0)


def test_canonicalize_add_neg():
    sa = cast(1.0)
    nsa = scalar.Negative(sa)

    outR = scalar.Add(sa, nsa)
    outL = scalar.Add(nsa, sa)

    outR = scalar.Scalar.canonicalize(outR)
    outL = scalar.Scalar.canonicalize(outL)

    assert outR == scalar.Literal(0.0)
    assert outL == scalar.Literal(0.0)

    A = cast(1.0)
    B = cast(2.0)

    C = scalar.Add(A, B)
    nC = -C

    assert scalar.Scalar.canonicalize(scalar.Add(C, nC)) == scalar.Literal(0.0)
    assert scalar.Scalar.canonicalize(scalar.Add(nC, C)) == scalar.Literal(0.0)


def test_canonicalize_mul_zero():
    zero = scalar.Literal(0.0)
    one = cast(1)

    outR = scalar.Mul(zero, one)
    outL = scalar.Mul(one, zero)

    outR = scalar.Scalar.canonicalize(outR)
    outL = scalar.Scalar.canonicalize(outL)

    assert outR == scalar.Literal(0.0)
    assert outL == scalar.Literal(0.0)

    A = cast(1.0)
    B = cast(2.0)

    C = scalar.Add(A, B)

    assert scalar.Scalar.canonicalize(scalar.Mul(zero, C)) == scalar.Literal(0.0)
    assert scalar.Scalar.canonicalize(scalar.Mul(C, zero)) == scalar.Literal(0.0)


def test_add_scalar():
    A = cast(1)
    B = cast(2)

    C = scalar.Add(A, B)

    assert C.children() == [A, B]
    assert C.print_node() == "+"
    assert str(C) == "(1 + 2)"


def test_add_zero():
    A = cast(1)
    B = cast(2)

    C = scalar.Add(A, B)  # expression
    zero = cast(0.0)

    LC = scalar.Scalar.canonicalize(scalar.Add(C, zero))
    RC = scalar.Scalar.canonicalize(scalar.Add(zero, C))

    assert LC == cast(3)
    assert RC == cast(3)


def test_mul_one():
    A = cast(2)
    B = cast(2)

    C = scalar.Add(A, B)  # expression
    one = cast(1.0)

    LC = scalar.Scalar.canonicalize(scalar.Mul(C, one))
    RC = scalar.Scalar.canonicalize(scalar.Mul(one, C))

    assert LC == cast(4)
    assert RC == cast(4)


def test_div_rone():
    A = cast(2)
    B = cast(2)

    C = scalar.Add(A, B)  # expression
    one = cast(1.0)

    LC = scalar.Scalar.canonicalize(scalar.Div(C, one))

    assert LC == cast(4)

    dC = A / B
    assert dC == cast(1)


def test_mul_scalar():
    A = cast(1)
    B = cast(2)

    C = scalar.Mul(A, B)

    assert C.children() == [A, B]
    assert C.print_node() == "*"


def test_div_scalar():
    A = cast(1)
    B = cast(2)

    C = scalar.Div(A, B)

    assert C.children() == [A, B]
    assert C.print_node() == "/"

    assert C() == 0.5


def test_min_scalar():
    A = cast(1)
    B = cast(2)
    C = cast(3)

    D = scalar.Min([A, B, C])
    assert D.children() == [A, B, C]
    assert D.print_node() == "min"

    mystdout = StringIO()
    p = PP(mystdout)
    D._repr_pretty_(p, 0)

    assert (
        mystdout.getvalue()
        == "min\n├─ Literal: 1\n⋮\n├─ Literal: 2\n⋮\n└─ Literal: 3⋮\n"
    )
    assert str(D) == "min(1, 2, 3)"


def test_max_scalar():
    A = cast(1)
    B = cast(2)
    C = cast(3)

    D = scalar.Max([A, B, C])
    assert D.children() == [A, B, C]
    assert D.print_node() == "max"

    mystdout = StringIO()
    p = PP(mystdout)
    D._repr_pretty_(p, 0)

    assert (
        mystdout.getvalue()
        == "max\n├─ Literal: 1\n⋮\n├─ Literal: 2\n⋮\n└─ Literal: 3⋮\n"
    )
    assert str(D) == "max(1, 2, 3)"


def test_cast_decimal():
    A = cast(Decimal("1.0"))

    assert A.value == Decimal("1.0")


def test_Interval_from_pyslice():
    with pytest.raises(ValueError):
        scalar.Interval.from_slice(slice(None, None, None))

    with pytest.raises(ValueError):
        scalar.Interval.from_slice(slice(None, 3, 3))

    with pytest.raises(ValueError):
        scalar.Interval.from_slice(slice(3, None, 3))

    with pytest.raises(ValueError):
        scalar.Interval.from_slice(slice(3, 3, 3))

    with pytest.raises(ValueError):
        scalar.Interval.from_slice(slice(None, None, 3))

    itvl = scalar.Interval.from_slice(slice(5, 6, None))
    assert itvl.start == cast(5)
    assert itvl.stop == cast(6)


def test_Slice():
    itvl = scalar.Interval(cast(5), cast(6))

    slc = scalar.Slice(cast(1), itvl)

    assert slc.children() == {"Scalar": cast(1), None: itvl}
    assert slc.print_node() == "Slice"

    mystdout = StringIO()
    p = PP(mystdout)

    slc._repr_pretty_(p, 0)

    assert (
        mystdout.getvalue()
        == "Slice\n"
        + "├─ Scalar\n"
        + "│  ⇒ Literal: 1\n"
        + "⋮\n"
        + "└─ Interval\n"
        + "⋮\n"
    )
    assert str(slc) == "(1)[5:6]"


def test_assigned_variable_methods():
    assigned_var = scalar.AssignedVariable("a", Decimal("1.0"))

    assert assigned_var.children() == []
    assert assigned_var.print_node() == "AssignedVariable: a = 1.0"



from bloqade.ir.routine.base import Routine
from bloqade.ir.control.waveform import Waveform, Linear, Constant
from bloqade.builder.typing import ScalarType
from beartype import beartype
from beartype.typing import TYPE_CHECKING, List, Optional, Union, Dict, Any
from decimal import Decimal

if TYPE_CHECKING:
    from bloqade.submission.ir.capabilities import QuEraCapabilities


def get_capabilities() -> "QuEraCapabilities":
    """Get the device capabilities for Aquila

    Returns:
        QuEraCapabilities: capabilities object for Aquila device.


    Note:
        Units of time, distance, and energy are microseconds (us),
        micrometers (um), and rad / us, respectively.

        For a comprehensive list of capabilities,
        see the [Hardware Reference](../../hardware-reference/hardware-capabilities.md)
        page
    """

    from bloqade.submission.capabilities import get_capabilities
    import bloqade.submission.ir.capabilities as cp

    capabilities_schema = get_capabilities()

    capabilities = capabilities_schema.capabilities

    # manually convert to units
    return cp.QuEraCapabilities(
        version=capabilities_schema.version,
        capabilities=cp.DeviceCapabilities(
            task=cp.TaskCapabilities(
                number_shots_max=capabilities.task.number_shots_max,
                number_shots_min=capabilities.task.number_shots_min,
            ),
            lattice=cp.LatticeCapabilities(
                number_qubits_max=capabilities.lattice.number_qubits_max,
                area=cp.LatticeAreaCapabilities(
                    width=(capabilities.lattice.area.width * Decimal("1e6")),  # m
                    height=(capabilities.lattice.area.height * Decimal("1e6")),  # m
                ),
                geometry=cp.LatticeGeometryCapabilities(
                    number_sites_max=capabilities.lattice.geometry.number_sites_max,
                    spacing_radial_min=(
                        capabilities.lattice.geometry.spacing_radial_min
                        * Decimal("1e6")
                    ),  # m
                    spacing_vertical_min=(
                        capabilities.lattice.geometry.spacing_vertical_min
                        * Decimal("1e6")
                    ),  # m
                    position_resolution=(
                        capabilities.lattice.geometry.position_resolution
                        * Decimal("1e6")
                    ),  # m
                ),
            ),
            rydberg=cp.RydbergCapabilities(
                c6_coefficient=capabilities.rydberg.c6_coefficient
                * Decimal("1e30"),  # rad * m^6 / s
                global_=cp.RydbergGlobalCapabilities(
                    phase_max=(capabilities.rydberg.global_.phase_max),  # rad
                    phase_min=(capabilities.rydberg.global_.phase_min),  # rad
                    phase_resolution=(
                        capabilities.rydberg.global_.phase_resolution
                    ),  # rad
                    rabi_frequency_max=(
                        capabilities.rydberg.global_.rabi_frequency_max / Decimal("1e6")
                    ),  # rad / s
                    rabi_frequency_min=(
                        capabilities.rydberg.global_.rabi_frequency_min / Decimal("1e6")
                    ),  # rad / s
                    rabi_frequency_resolution=(
                        capabilities.rydberg.global_.rabi_frequency_resolution
                        / Decimal("1e6")
                    ),  # rad / s
                    rabi_frequency_slew_rate_max=(
                        capabilities.rydberg.global_.rabi_frequency_slew_rate_max
                        / Decimal("1e6") ** 2
                    ),  # rad / s^2
                    detuning_max=(
                        capabilities.rydberg.global_.detuning_max / Decimal("1e6")
                    ),  # rad / s
                    detuning_min=(
                        capabilities.rydberg.global_.detuning_min / Decimal("1e6")
                    ),  # rad / s
                    detuning_resolution=(
                        capabilities.rydberg.global_.detuning_resolution
                        / Decimal("1e6")
                    ),  # rad / s
                    detuning_slew_rate_max=(
                        capabilities.rydberg.global_.detuning_slew_rate_max
                        / Decimal("1e6") ** 2
                    ),  # rad / s^2
                    time_min=(
                        capabilities.rydberg.global_.time_min * Decimal("1e6")
                    ),  # s
                    time_max=(
                        capabilities.rydberg.global_.time_max * Decimal("1e6")
                    ),  # s
                    time_resolution=(
                        capabilities.rydberg.global_.time_resolution * Decimal("1e6")
                    ),  # s
                    time_delta_min=(
                        capabilities.rydberg.global_.time_delta_min * Decimal("1e6")
                    ),  # s
                ),
                local=cp.RydbergLocalCapabilities(
                    number_local_detuning_sites=(
                        capabilities.rydberg.local.number_local_detuning_sites
                    ),
                    site_coefficient_max=(
                        capabilities.rydberg.local.site_coefficient_max
                    ),
                    site_coefficient_min=(
                        capabilities.rydberg.local.site_coefficient_min
                    ),
                    spacing_radial_min=(
                        capabilities.rydberg.local.spacing_radial_min * Decimal("1e6")
                    ),  # rad / s
                    detuning_min=(
                        capabilities.rydberg.local.detuning_min / Decimal("1e6")
                    ),  # rad / s
                    detuning_max=(
                        capabilities.rydberg.local.detuning_max / Decimal("1e6")
                    ),  # rad / s
                    detuning_slew_rate_max=(
                        capabilities.rydberg.local.detuning_slew_rate_max
                        / Decimal("1e6") ** 2
                    ),  # rad / s^2
                    time_delta_min=(
                        capabilities.rydberg.local.time_delta_min * Decimal("1e6")
                    ),  # s
                    time_resolution=(
                        capabilities.rydberg.local.time_resolution * Decimal("1e6")
                    ),  # s
                ),
            ),
        ),
    )


@beartype
def linear(duration: ScalarType, start: ScalarType, stop: ScalarType) -> Linear:
    """Create a Linear waveform.

    Args:
        duration (ScalarType): Duration of linear waveform
        start (ScalarType): Starting value of linear waveform
        stop (ScalarType): Ending value of linear waveform

    Returns:
        Linear: Linear waveform
    """
    return Linear(start, stop, duration)


@beartype
def constant(duration: ScalarType, value: ScalarType) -> Constant:
    """Create a Constant waveform.

    Args:
        duration (ScalarType): Duration of the Constant waveform.
        value (ScalarType): Value of the Constant waveform.s

    Returns:
        Constant: A Constant waveform.
    """
    return Constant(value, duration)


@beartype
def piecewise_linear(durations: List[ScalarType], values: List[ScalarType]) -> Waveform:
    """Create a piecewise linear waveform.

    Create a piecewise linear waveform from a list of durations and values. The
    value `duration[i]` is of the linear segment between `values[i]` and `values[i+1]`.

    Args:
        durations (List[ScalarType]): The duration of each segment
        values (List[ScalarType]): The values for each segment

    Raises:
        ValueError: If the length of `values` is not one greater than the length of
        `durations`.

    Returns:
        Waveform: The piecewise linear waveform.
    """

    if len(durations) + 1 != len(values):
        raise ValueError(
            "The length of values must be one greater than the length of durations"
        )

    pwl_wf = None
    for duration, start, stop in zip(durations, values[:-1], values[1:]):
        if pwl_wf is None:
            pwl_wf = Linear(start, stop, duration)
        else:
            pwl_wf = pwl_wf.append(Linear(start, stop, duration))

    return pwl_wf


@beartype
def piecewise_constant(
    durations: List[ScalarType], values: List[ScalarType]
) -> Waveform:
    """Create a piecewise linear waveform.

    Create a piecewise constant waveform from a list of durations and values. The
    value `duration[i]` corresponds to the length of time for the i'th segment
    with a value of `values[i]`.

    Args:
        durations (List[ScalarType]): The duration of each segment
        values (List[ScalarType]): The values for each segment

    Raises:
        ValueError: If the length of `values` is not the same as the length of
        `durations`.

    Returns:
        Waveform: The piecewise linear waveform.
    """
    if len(durations) != len(values):
        raise ValueError(
            "The length of values must be the same as the length of durations"
        )

    pwc_wf = None
    for duration, value in zip(durations, values):
        if pwc_wf is None:
            pwc_wf = Constant(value, duration)
        else:
            pwc_wf = pwc_wf.append(Constant(value, duration))

    return pwc_wf


@beartype
def rydberg_h(
    atoms_positions: Any,
    detuning: Optional[Waveform] = None,
    amplitude: Optional[Waveform] = None,
    phase: Optional[Waveform] = None,
    static_params: Dict[str, Any] = {},
    batch_params: Union[List[Dict[str, Any]], Dict[str, Any]] = [],
    args: List[str] = [],
) -> Routine:
    """Create a rydberg program with uniform detuning, amplitude, and phase.

    Args:
        atoms_positions (Any): Description of geometry of atoms in system.
        detuning (Optional[Waveform], optional): Waveform for detuning.
            Defaults to None.
        amplitude (Optional[Waveform], optional): Waveform describing the amplitude of
            the rabi term. Defaults to None.
        phase (Optional[Waveform], optional): Waveform describing the phase of rabi
            term. Defaults to None.
        static_params (Dict[str, Any], optional): Define static parameters of your
            program. Defaults to {}.
        batch_params (Union[List[Dict[str, Any]], Dict[str, Any]], optional):
            Parmaters for a batch of tasks. Defaults to [].
        args (List[str], optional): List of arguments to leave till runtime.
            Defaults to [].

    Returns:
        Routine: An object that can be used to dispatch a rydberg program to
            multiple backends.
    """
    from bloqade import start
    from bloqade.atom_arrangement import AtomArrangement

    if isinstance(atoms_positions, AtomArrangement):
        prog = atoms_positions
    else:
        prog = start.add_position(atoms_positions)

    if detuning is not None:
        prog = prog.rydberg.detuning.uniform.apply(detuning)

    if amplitude is not None:
        prog = prog.amplitude.uniform.apply(amplitude)

    if phase is not None:
        prog = prog.phase.uniform.apply(phase)

    prog = prog.assign(**static_params)

    if isinstance(batch_params, dict):
        prog = prog.batch_assign(**batch_params)
    else:
        prog = prog.batch_assign(batch_params)

    prog = prog.args(args)

    return prog.parse()



from bloqade.ir import (
    Linear,
    Constant,
    Poly,
    Record,
    AlignedWaveform,
    Alignment,
    AlignedValue,
    to_waveform,
    Interpolation,
    GaussianKernel,
    LogisticKernel,
    SigmoidKernel,
    TriangleKernel,
    UniformKernel,
    ParabolicKernel,
    BiweightKernel,
    TriweightKernel,
    TricubeKernel,
    CosineKernel,
)
from bloqade import cast
from bloqade.ir.scalar import Interval
from bloqade.ir.control.waveform import PythonFn, Append, Slice, Sample
from bloqade.ir.control.waveform import SmoothingKernel, Waveform
from decimal import Decimal
import pytest
import numpy as np
from io import StringIO
from IPython.lib.pretty import PrettyPrinter as PP
import bloqade.ir.tree_print as trp

trp.color_enabled = False


def test_wvfm_base():
    wf = Waveform()
    with pytest.raises(NotImplementedError):
        wf.eval_decimal(Decimal("0.5"))


def test_wvfm_linear():
    wf = Linear(start=1.0, stop=2.0, duration=3.0)

    assert wf.print_node() == "Linear"
    assert wf.eval_decimal(clock_s=Decimal("6.0")) == 0
    assert isinstance(hash(wf), int)

    assert wf.children() == {
        "start": cast(1.0),
        "stop": cast(2.0),
        "duration": cast(3.0),
    }


def test_wvfm_constant():
    wf = Constant(value=1.0, duration=3.0)

    assert wf.print_node() == "Constant"
    assert wf.eval_decimal(clock_s=Decimal("6.0")) == 0
    assert wf.children() == {"value": cast(1.0), "duration": cast(3.0)}
    assert isinstance(hash(wf), int)

    mystdout = StringIO()
    p = PP(mystdout)

    wf._repr_pretty_(p, 0)

    assert (
        mystdout.getvalue()
        == "Constant\n"
        + "├─ value\n"
        + "│  ⇒ Literal: 1.0\n"
        + "⋮\n"
        + "└─ duration\n"
        + "   ⇒ Literal: 3.0⋮\n"
    )


def test_wvfm_pyfn():
    def my_func(time, *, omega, phi=0, amplitude):
        import numpy as np

        return amplitude * np.cos(omega * time + phi)

    @to_waveform(duration=1.0)
    def annot_my_func(time, *, omega, phi=0, amplitude):
        import numpy as np

        return amplitude * np.cos(omega * time + phi)

    ## have varargs:
    def my_func2(time, *omega):
        return time

    assert my_func2(3) == 3

    ## have varkw:
    def my_func3(time, omega, **phi):
        return time

    assert my_func3(3, 2) == 3

    with pytest.raises(ValueError):
        PythonFn.create(my_func2, duration=1.0)

    with pytest.raises(ValueError):
        PythonFn.create(my_func3, duration=1.0)

    wf = PythonFn.create(my_func, duration=1.0)
    awf = annot_my_func

    assert wf.eval_decimal(Decimal("0.56"), omega=1, amplitude=4) == awf.eval_decimal(
        Decimal("0.56"), omega=1, amplitude=4
    )

    assert wf.eval_decimal(Decimal("3"), omega=1, amplitude=4) == Decimal(0)

    assert wf.print_node() == "PythonFn: my_func"
    assert wf.children() == {
        "duration": cast(1.0),
        "omega": cast("omega"),
        "phi": cast("phi"),
        "amplitude": cast("amplitude"),
    }
    assert wf.duration == cast(1.0)
    assert isinstance(hash(wf), int)

    mystdout = StringIO()
    p = PP(mystdout)

    wf._repr_pretty_(p, 0)

    assert mystdout.getvalue() == (
        "PythonFn: my_func\n"
        + "├─ duration\n"
        + "│  ⇒ Literal: 1.0\n⋮\n"
        + "├─ phi\n"
        + "│  ⇒ Variable: phi\n⋮\n"
        + "├─ omega\n"
        + "│  ⇒ Variable: omega\n⋮\n"
        + "└─ amplitude\n"
        + "   ⇒ Variable: amplitude⋮\n"
    )


def test_wvfm_app():
    wf = Linear(start=1.0, stop=2.0, duration=3.0)
    wf2 = Constant(value=1.0, duration=3.0)

    wf3 = Append([wf, wf2])

    assert wf3.print_node() == "Append"
    assert wf3.children() == (wf, wf2)
    assert wf3.eval_decimal(Decimal(10)) == Decimal(0)
    assert isinstance(hash(wf), int)

    mystdout = StringIO()
    p = PP(mystdout)

    wf3._repr_pretty_(p, 0)

    assert (
        mystdout.getvalue()
        == "Append\n" + "├─ Linear\n" + "⋮\n" + "└─ Constant\n" + "⋮\n"
    )


def test_wvfm_neg():
    wf = Constant(value=1.0, duration=3.0)
    wf2 = -wf

    assert wf2.print_node() == "Negative"
    assert wf2.children() == [wf]
    assert isinstance(hash(wf), int)

    assert wf2.eval_decimal(Decimal("0.5")) == Decimal("-1.0")

    mystdout = StringIO()
    p = PP(mystdout)

    wf2._repr_pretty_(p, 2)

    assert (
        mystdout.getvalue()
        == "Negative\n"
        + "└─ Constant\n"
        + "   ├─ value\n"
        + "   │  ⇒ Literal: 1.0\n"
        + "   └─ duration\n"
        + "      ⇒ Literal: 3.0"
    )


def test_wvfm_scale():
    wf = Constant(value=1.0, duration=3.0)
    wf2 = 2.0 * wf

    assert wf2.print_node() == "Scale"
    assert wf2.children() == [cast(2.0), wf]
    assert isinstance(hash(wf), int)

    assert wf2.eval_decimal(Decimal("0.5")) == Decimal("2.0")

    wf3 = wf * 2.0
    assert wf3.print_node() == "Scale"
    assert wf3.children() == [cast(2.0), wf]

    assert wf3.eval_decimal(Decimal("0.5")) == Decimal("2.0")

    mystdout = StringIO()
    p = PP(mystdout)

    wf3._repr_pretty_(p, 2)

    assert (
        mystdout.getvalue()
        == "Scale\n"
        + "├─ Literal: 2.0\n"
        + "└─ Constant\n"
        + "   ├─ value\n"
        + "   │  ⇒ Literal: 1.0\n"
        + "   └─ duration\n"
        + "      ⇒ Literal: 3.0"
    )


def test_wvfn_add():
    wf = Constant(value=1.0, duration=3.0)
    wf2 = Linear(start=1.0, stop=2.0, duration=2.0)

    wf3 = wf + wf2

    assert wf3.print_node() == "+"
    assert wf3.children() == [wf, wf2]
    assert isinstance(hash(wf), int)

    assert wf3.eval_decimal(Decimal("0")) == Decimal("2.0")
    assert wf3.eval_decimal(Decimal("2.5")) == Decimal("1.0")

    mystdout = StringIO()
    p = PP(mystdout)

    wf3._repr_pretty_(p, 2)

    assert (
        mystdout.getvalue()
        == "+\n"
        + "├─ Constant\n"
        + "│  ├─ value\n"
        + "│  │  ⇒ Literal: 1.0\n"
        + "│  └─ duration\n"
        + "│     ⇒ Literal: 3.0\n"
        + "└─ Linear\n"
        + "   ├─ start\n"
        + "   │  ⇒ Literal: 1.0\n"
        + "   ├─ stop\n"
        + "   │  ⇒ Literal: 2.0\n"
        + "   └─ duration\n"
        + "      ⇒ Literal: 2.0"
    )


def test_wvfn_rec():
    wf = Linear(start=1.0, stop=2.0, duration=3.0)

    re = Record(wf, cast("tst"))

    assert re.print_node() == "Record"
    assert re.children() == {"Waveform": wf, "Variable": cast("tst")}
    assert isinstance(hash(wf), int)

    assert re.eval_decimal(Decimal("0")) == Decimal("1.0")
    assert re.duration == cast(3.0)

    mystdout = StringIO()
    p = PP(mystdout)

    re._repr_pretty_(p, 2)

    assert (
        mystdout.getvalue()
        == "Record\n"
        + "├─ Waveform\n"
        + "│  ⇒ Linear\n"
        + "│    ├─ start\n"
        + "│    │  ⇒ Literal: 1.0\n"
        + "│    ├─ stop\n"
        + "│    │  ⇒ Literal: 2.0\n"
        + "│    └─ duration\n"
        + "│       ⇒ Literal: 3.0\n"
        + "└─ Variable\n"
        + "   ⇒ Variable: tst"
    )


def test_wvfn_poly():
    wf = Poly(coeffs=[cast(1), cast(2), cast(3)], duration=10)

    assert wf.print_node() == "Poly"
    assert wf.children() == {
        "b": wf.coeffs[0],
        "t": wf.coeffs[1],
        "t^2": wf.coeffs[2],
        "duration": cast(10),
    }
    assert wf.eval_decimal(Decimal("0.5")) == (1) + (2) * 0.5 + (3) * 0.5**2
    assert wf.eval_decimal(Decimal("20")) == Decimal("0")
    assert isinstance(hash(wf), int)


##-----------------------------


def test_smkern_base():
    ker = SmoothingKernel()

    with pytest.raises(NotImplementedError):
        ker(0.5)


def test_smkern_gauss():
    ker = GaussianKernel
    assert ker(0.5) == np.exp(-(0.5**2) / 2) / np.sqrt(2 * np.pi)


def test_smkern_logist():
    ker = LogisticKernel
    assert ker(0.5) == np.exp(-(np.logaddexp(0, 0.5) + np.logaddexp(0, -0.5)))


def test_smkern_sigmoid():
    ker = SigmoidKernel
    assert ker(0.5) == (2 / np.pi) * np.exp(-np.logaddexp(-0.5, 0.5))


def test_smkern_triangle():
    ker = TriangleKernel
    assert ker(0.5) == np.maximum(0, 1 - np.abs(0.5))


def test_smkern_uniform():
    ker = UniformKernel
    assert ker(0.5) == 1.0


def test_smkern_parab():
    ker = ParabolicKernel
    assert ker(0.5) == (3 / 4) * np.maximum(0, 1 - 0.5**2)


def test_smkern_biweight():
    ker = BiweightKernel
    assert ker(0.5) == (15 / 16) * np.maximum(0, 1 - 0.5**2) ** 2


def test_smkern_triweight():
    ker = TriweightKernel
    assert ker(0.5) == (35 / 32) * np.maximum(0, 1 - 0.5**2) ** 3


def test_smkern_tricube():
    ker = TricubeKernel
    assert ker(0.5) == (70 / 81) * np.maximum(0, 1 - np.abs(0.5) ** 3) ** 3


def test_smkern_cosine():
    ker = CosineKernel
    assert ker(0.5) == (np.pi / 4) * np.cos(np.pi * 0.5 / 2)


def test_wvfn_smooth():
    wv = Linear(start=1.0, stop=2.0, duration=3.0)
    wf = wv.smooth(radius=0.5, kernel=GaussianKernel)

    assert wf.duration == cast(3.0)

    assert wf.eval_decimal(Decimal("0.1")) == 1.0844831620655968


def test_wvfn_slice():
    iv = Interval(cast(0), cast(0.3))
    wv = Constant(value=2.0, duration=3.0)

    wf = Slice(wv, iv)

    assert wf.print_node() == "Slice"
    assert wf.eval_decimal(Decimal("0.4")) == 0
    assert wf.eval_decimal(Decimal("0.2")) == 2.0
    assert wf.children() == [wv, iv]
    assert isinstance(hash(wf), int)

    mystdout = StringIO()
    p = PP(mystdout)
    wf._repr_pretty_(p, 2)

    assert (
        mystdout.getvalue()
        == "Slice\n"
        + "├─ Constant\n"
        + "│  ├─ value\n"
        + "│  │  ⇒ Literal: 2.0\n"
        + "│  └─ duration\n"
        + "│     ⇒ Literal: 3.0\n"
        + "└─ Interval\n"
        + "   ├─ start\n"
        + "   │  ⇒ Literal: 0\n"
        + "   └─ stop\n"
        + "      ⇒ Literal: 0.3"
    )

    iv_err1 = Interval(None, None)
    wf2 = Slice(wv, iv_err1)
    with pytest.raises(ValueError):
        wf2.duration

    iv2 = Interval(cast(0), None)
    wf3 = Slice(wv, iv2)

    assert wf3.duration() == Decimal("3.0")


def test_wvfm_align():
    wv = Constant(value=2.0, duration=3.0)

    wf = AlignedWaveform(wv, Alignment.Left, cast(0.2))
    assert wf.print_node() == "AlignedWaveform"
    assert wf.children() == {"Waveform": wv, "Alignment": "Left", "Value": cast(0.2)}
    assert isinstance(hash(wf), int)

    wf2 = AlignedWaveform(wv, Alignment.Left, AlignedValue.Right)
    assert wf2.print_node() == "AlignedWaveform"
    assert wf2.children() == {"Waveform": wv, "Alignment": "Left", "Value": "Right"}

    wf3 = AlignedWaveform(wv, Alignment.Right, AlignedValue.Left)
    assert wf3.print_node() == "AlignedWaveform"
    assert wf3.children() == {"Waveform": wv, "Alignment": "Right", "Value": "Left"}

    mystdout = StringIO()
    p = PP(mystdout)
    wf3._repr_pretty_(p, 2)

    assert (
        mystdout.getvalue()
        == "AlignedWaveform\n"
        + "├─ Waveform\n"
        + "│  ⇒ Constant\n"
        + "│    ├─ value\n"
        + "│    │  ⇒ Literal: 2.0\n"
        + "│    └─ duration\n"
        + "│       ⇒ Literal: 3.0\n"
        + "├─ Alignment\n"
        + "│  ⇒ Right\n"
        + "└─ Value\n"
        + "   ⇒ Left\n"
    )


def test_wvfm_sample():
    def my_cos(time):
        return np.cos(time)

    assert my_cos(1) == np.cos(1)

    wv = PythonFn.create(my_cos, duration=1.0)
    dt = cast(0.1)

    wf = Sample(wv, Interpolation.Constant, dt)

    assert wf.print_node() == "Sample constant"
    assert wf.children() == {"Waveform": wv, "sample_step": dt}
    assert wf.eval_decimal(Decimal(0.05)) == my_cos(0)
    assert float(wf.eval_decimal(Decimal(0))) == my_cos(0)
    assert isinstance(hash(wf), int)

    wf2 = Sample(wv, Interpolation.Linear, dt)

    assert wf2.print_node() == "Sample linear"
    assert wf2.children() == {"Waveform": wv, "sample_step": dt}
    slope = (my_cos(0.1) - my_cos(0)) / 0.1
    assert float(wf2.eval_decimal(Decimal(0.05))) == float(my_cos(0) + slope * 0.05)
    assert float(wf2.eval_decimal(Decimal(3))) == 0
    assert float(wf2.eval_decimal(Decimal(0))) == my_cos(0)
    assert isinstance(hash(wf), int)

    mystdout = StringIO()
    p = PP(mystdout)

    wf2._repr_pretty_(p, 2)

    assert (
        mystdout.getvalue()
        == "Sample linear\n"
        + "├─ Waveform\n"
        + "│  ⇒ PythonFn: my_cos\n"
        + "│    └─ duration\n"
        + "│       ⇒ Literal: 1.0\n"
        + "└─ sample_step\n"
        + "   ⇒ Literal: 0.1"
    )


"""
print(wf[:0.5].duration)
print(wf[1.0:].duration)
print(wf[0.2:0.8].duration)

print(-wf)
print(wf.scale(1.0))

eval(repr(-wf))
eval(repr(wf.scale(1.0)))

# canonicalize append
wf = (
    Linear(0.0, "rabi_amplitude_max", "up_time")
    .append(Constant("rabi_amplitude_max", "anneal_time"))
    .append(Linear("rabi_amplitude_max", 0.0, "up_time"))
)
print(wf)

smooth_wf = wf.smooth(0.1, GaussianKernel)
print(smooth_wf(1.0, rabi_amplitude_max=1.0, up_time=1.0, anneal_time=1.0))
smooth_wf = wf.smooth(0.1, BiweightKernel)
print(smooth_wf(1.0, rabi_amplitude_max=1.0, up_time=1.0, anneal_time=1.0))

# try scaling
wf * scalar.Literal(5.0)

# try addition
wf + wf

# polynomial
Poly(
    [scalar.Literal(10) + scalar.Variable("l"), scalar.Literal(5), scalar.Literal(-2)],
    scalar.Variable("g"),
)

# Record
Record(wf, scalar.Variable("n"))

AlignedWaveform(wf, Alignment.Left, AlignedValue.Right)
"""
# eval(repr(wf))

# print(wf)

# wf = wf.append(wf)
# eval(repr(wf))
# print(wf)

# wf = Linear(0.0, "rabi_amplitude_max", "up_time").append(wf)
# eval(repr(wf))
# print(wf)



from bloqade import start, cast
from bloqade.ir.location import Square, Rectangular
import random
import numpy as np
import os
import bloqade.ir.tree_print as trp


trp.color_enabled = False


PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH = os.path.join(
    os.getcwd(), "tests/data/expected_pprint_output"
)


def test_list_of_locations_pprint():
    random.seed(1337)
    rand_positions = list(
        zip(
            [random.randint(0, 20) for _ in range(10)],
            [random.randint(0, 20) for _ in range(10)],
        )
    )

    list_of_locations_pprint_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH, "list_of_locations_pprint_output.txt"
    )
    list_of_locations_pprint_output = open(
        list_of_locations_pprint_output_path, "r"
    ).read()

    assert str(start.add_position(rand_positions)) == list_of_locations_pprint_output

    var1 = cast("var1")
    var2 = cast("var2")
    var3 = cast("var3")

    variable_list_of_locations = start.add_position([(var1, var2), (var3, 5), (0, 9)])

    assert str(variable_list_of_locations) == (
        "AtomArrangement\n"
        "├─ Location: filled\n"
        "│  ├─ x\n"
        "│  │  ⇒ Variable: var1\n"
        "│  └─ y\n"
        "│     ⇒ Variable: var2\n"
        "├─ Location: filled\n"
        "│  ├─ x\n"
        "│  │  ⇒ Variable: var3\n"
        "│  └─ y\n"
        "│     ⇒ Literal: 5\n"
        "└─ Location: filled\n"
        "   ├─ x\n"
        "   │  ⇒ Literal: 0\n"
        "   └─ y\n"
        "      ⇒ Literal: 9"
    )


def test_square_pprint():
    # full
    square_pprint_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH, "square_pprint_output.txt"
    )
    square_pprint_output = open(square_pprint_output_path, "r").read()

    assert str(Square(7)) == square_pprint_output

    # apply defect count
    square_pprint_defect_count_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH,
        "square_pprint_defect_count_output.txt",
    )
    square_pprint_defect_count_output = open(
        square_pprint_defect_count_output_path, "r"
    ).read()
    assert (
        str(Square(7).apply_defect_count(21, np.random.default_rng(1337)))
        == square_pprint_defect_count_output
    )

    # apply defect density
    square_pprint_defect_density_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH,
        "square_pprint_defect_density_output.txt",
    )
    square_pprint_defect_density_output = open(
        square_pprint_defect_density_output_path, "r"
    ).read()

    assert (
        str(Square(7).apply_defect_density(0.5, np.random.default_rng(1337)))
        == square_pprint_defect_density_output
    )

    # apply variable for lattice
    square_pprint_var_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH, "square_pprint_var_output.txt"
    )
    square_pprint_var_output = open(square_pprint_var_output_path, "r").read()

    bl = cast("bl")
    assert str(Square(7, lattice_spacing=bl)) == square_pprint_var_output


def test_rectangular_pprint():
    rectangular_pprint_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH, "rectangular_pprint_output.txt"
    )
    rectangular_pprint_output = open(rectangular_pprint_output_path, "r").read()

    assert str(Rectangular(7, 5)) == rectangular_pprint_output

    rectangular_pprint_defect_count_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH,
        "rectangular_pprint_defect_count_output.txt",
    )
    rectangular_pprint_defect_count_output = open(
        rectangular_pprint_defect_count_output_path, "r"
    ).read()

    assert (
        str(Rectangular(7, 5).apply_defect_count(15, np.random.default_rng(1337)))
        == rectangular_pprint_defect_count_output
    )

    rectangular_pprint_defect_density_output_path = os.path.join(
        PROJECT_RELATIVE_PPRINT_TESTS_OUTPUT_PATH,
        "rectangular_pprint_defect_density_output.txt",
    )
    rectangular_pprint_defect_density_output = open(
        rectangular_pprint_defect_density_output_path, "r"
    ).read()

    assert (
        str(Rectangular(7, 5).apply_defect_density(0.5, np.random.default_rng(1337)))
        == rectangular_pprint_defect_density_output
    )



from bloqade.ir.scalar import Scalar
from decimal import Decimal
from numbers import Real
from beartype.typing import List, Union


ScalarType = Union[Real, Decimal, str, Scalar]
LiteralType = Union[Real, Decimal]
ParamType = Union[LiteralType, List[LiteralType]]



from bloqade.atom_arrangement import Chain
from bloqade.ir.analysis.is_constant import IsConstantAnalogCircuit


def test_happy_path():
    circuit = (
        Chain(8, 6.1)
        .rydberg.detuning.uniform.constant(1.0, 10.0)
        .amplitude.uniform.linear(1.0, 1.0, 5.0)
        .linear(1.0, 1.0, 5.0)
        .phase.uniform.poly([1.0, 0.0, 0.0, 0.0, 0.0], 10)
        .parse_circuit()
    )

    result = IsConstantAnalogCircuit().emit(circuit)
    assert result.is_constant

    circuit = (
        Chain(8, 6.1)
        .rydberg.detuning.uniform.constant(1.0, 5.0)
        .uniform.constant(1.0, 10.0)
        .slice(2.5, 7.5)
        .amplitude.uniform.linear(1.0, 1.0, 5.0)
        .linear(1.0, 1.0, 5.0)
        .slice(2.5, 7.5)
        .record("var")
        .phase.uniform.poly([1.0, 0.0, 0.0, 0.0, 0.0], 5)
        .parse_circuit()
    )
    print(circuit)
    # assert False
    result = IsConstantAnalogCircuit().emit(circuit)
    assert result.is_constant


def test_fail_shape():
    circuit = (
        Chain(8, 6.1)
        .rydberg.detuning.uniform.constant(1.0, 5.0)
        .uniform.constant(1.0, 10.0)
        .slice(2.5, 7.5)
        .amplitude.uniform.linear(1.0, 1.0, 5.0)
        .linear(1.0, 1.0, 5.0)
        .slice(2.5, 7.5)
        .record("var")
        .phase.uniform.poly([1.0, 1.0, 0.0, 0.0, 0.0], 5)
        .parse_circuit()
    )
    print(circuit)
    # assert False
    result = IsConstantAnalogCircuit().emit(circuit)
    assert not result.is_constant

    circuit = (
        Chain(8, 6.1)
        .rydberg.detuning.uniform.constant(1.0, 5.0)
        .uniform.constant(1.0, 10.0)
        .slice(2.5, 7.5)
        .amplitude.uniform.linear(1.0, 1.0, 5.0)
        .linear(1.0, 1.0, 5.0)
        .slice(2.5, 7.5)
        .record("var")
        .phase.uniform.poly([1.0, 1.0, 0.0, 0.0, 0.0], 5)
        .amplitude.uniform.fn(lambda t: t, 5)
        .sample(0.05)
        .parse_circuit()
    )
    print(circuit)
    # assert False
    result = IsConstantAnalogCircuit().emit(circuit)
    assert not result.is_constant


def test_fail_duration():
    circuit = (
        Chain(8, 6.1)
        .rydberg.detuning.uniform.constant(1.0, 9.0)
        .amplitude.uniform.linear(1.0, 1.0, 5.0)
        .linear(1.0, 1.0, 5.0)
        .phase.uniform.poly([1.0, 0.0, 0.0, 0.0, 0.0], 10)
        .parse_circuit()
    )

    print(circuit)
    # assert False
    result = IsConstantAnalogCircuit().emit(circuit)
    assert not result.is_constant

    circuit = (
        Chain(8, 6.1)
        .rydberg.detuning.uniform.constant(1.0, 10.0)
        .uniform.constant(1.0, 9.0)
        .amplitude.uniform.linear(1.0, 1.0, 5.0)
        .linear(1.0, 1.0, 5.0)
        .phase.uniform.poly([1.0, 0.0, 0.0, 0.0, 0.0], 10)
        .parse_circuit()
    )

    print(circuit)
    # assert False
    result = IsConstantAnalogCircuit().emit(circuit)
    assert not result.is_constant


def test_fail_value():
    circuit = (
        Chain(8, 6.1)
        .rydberg.detuning.uniform.constant(1.0, 10.0)
        .constant(1.1, 10.0)
        .parse_circuit()
    )

    print(circuit)
    # assert False
    result = IsConstantAnalogCircuit().emit(circuit)
    assert not result.is_constant



import bloqade.ir.control.waveform as waveform
import bloqade.ir.control.field as field
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.sequence as sequence
import bloqade.ir.location.base as location
import bloqade.ir.analog_circuit as analog_circuit

from bloqade.ir.control.waveform import Waveform
from bloqade.ir.control.field import Field, SpatialModulation
from bloqade.ir.control.pulse import PulseExpr
from bloqade.ir.control.sequence import SequenceExpr
from bloqade.ir.location.base import AtomArrangement, ParallelRegister
from bloqade.ir.analog_circuit import AnalogCircuit
from typing import Union, Any


AstType = Union[
    Waveform,
    Field,
    SpatialModulation,
    PulseExpr,
    SequenceExpr,
    AtomArrangement,
    ParallelRegister,
    AnalogCircuit,
]


class AnalogCircuitVisitor:
    def visit_waveform(self, ast: waveform.Waveform) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__}"
            " for waveform.Waveform"
        )

    def visit_field(self, ast: field.Field) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} " "for field.Field"
        )

    def visit_uniform_modulation(self, ast: field.UniformModulation) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} "
            "for field.UniformModulation"
        )

    def visit_location(self, ast: field.Location) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} " "for field.Location"
        )

    def visit_scaled_locations(self, ast: field.ScaledLocations) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} "
            "for field.ScaledLocations"
        )

    def visit_run_time_vector(self, ast: field.RunTimeVector) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} "
            "for field.RunTimeVector"
        )

    def visit_assigned_run_time_vector(self, ast: field.AssignedRunTimeVector) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} "
            "for field.AssignedRunTimeVector"
        )

    def visit_pulse(self, ast: pulse.Pulse) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} " "for pulse.Pulse"
        )

    def visit_named_pulse(self, ast: pulse.NamedPulse) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} " "for pulse.NamedPulse"
        )

    def visit_append_pulse(self, ast: pulse.Append) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} " "for pulse.Append"
        )

    def visit_slice_pulse(self, ast: pulse.Slice) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} " "for pulse.Slice"
        )

    def visit_sequence(self, ast: sequence.Sequence) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} "
            "for sequence.Sequence"
        )

    def visit_named_sequence(self, ast: sequence.NamedSequence) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} "
            "for sequence.NamedSequence"
        )

    def visit_append_sequence(self, ast: sequence.Append) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} " "for sequence.Append"
        )

    def visit_slice_sequence(self, ast: sequence.Slice) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} " "for sequence.Slice"
        )

    def visit_register(self, ast: location.AtomArrangement) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} "
            "for location.AtomArrangement"
        )

    def visit_parallel_register(self, ast: location.ParallelRegister) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} "
            "for MultuplexRegister"
        )

    def visit_analog_circuit(self, ast: analog_circuit.AnalogCircuit) -> Any:
        raise NotImplementedError(
            f"No visitor method implemented in {self.__class__} "
            "for analog_circuit.AnalogCircuit"
        )

    def visit(self, ast: AstType) -> Any:
        if isinstance(ast, waveform.Waveform):
            return self.visit_waveform(ast)
        elif isinstance(ast, field.UniformModulation):
            return self.visit_uniform_modulation(ast)
        elif isinstance(ast, field.Location):
            return self.visit_location(ast)
        elif isinstance(ast, field.ScaledLocations):
            return self.visit_scaled_locations(ast)
        elif isinstance(ast, field.RunTimeVector):
            return self.visit_run_time_vector(ast)
        elif isinstance(ast, field.AssignedRunTimeVector):
            return self.visit_assigned_run_time_vector(ast)
        elif isinstance(ast, field.Field):
            return self.visit_field(ast)
        elif isinstance(ast, pulse.Pulse):
            return self.visit_pulse(ast)
        elif isinstance(ast, pulse.NamedPulse):
            return self.visit_named_pulse(ast)
        elif isinstance(ast, pulse.Append):
            return self.visit_append_pulse(ast)
        elif isinstance(ast, pulse.Slice):
            return self.visit_slice_pulse(ast)
        elif isinstance(ast, sequence.Sequence):
            return self.visit_sequence(ast)
        elif isinstance(ast, sequence.NamedSequence):
            return self.visit_named_sequence(ast)
        elif isinstance(ast, sequence.Append):
            return self.visit_append_sequence(ast)
        elif isinstance(ast, sequence.Slice):
            return self.visit_slice_sequence(ast)
        elif isinstance(ast, location.AtomArrangement):
            return self.visit_register(ast)
        elif isinstance(ast, location.ParallelRegister):
            return self.visit_parallel_register(ast)
        elif isinstance(ast, analog_circuit.AnalogCircuit):
            return self.visit_analog_circuit(ast)
        else:
            raise NotImplementedError(
                f"No visitor method implemented in {self.__class__} for {type(ast)}"
            )



from pydantic import BaseModel
from typing import Optional, List, Tuple
from decimal import Decimal
from bloqade.submission.ir.capabilities import QuEraCapabilities
from bloqade.visualization import display_task_ir, get_task_ir_figure

__all__ = ["QuEraTaskSpecification"]

# TODO: add version to these models.

FloatType = Decimal


def discretize_list(list_of_values: list, resolution: FloatType):
    resolution = Decimal(str(float(resolution)))
    return [round(Decimal(value) / resolution) * resolution for value in list_of_values]


class GlobalField(BaseModel):
    times: List[FloatType]
    values: List[FloatType]

    def __hash__(self):
        return hash((GlobalField, tuple(self.times), tuple(self.values)))


class LocalField(BaseModel):
    times: List[FloatType]
    values: List[FloatType]
    lattice_site_coefficients: List[FloatType]

    def __hash__(self):
        return hash(
            (
                LocalField,
                tuple(self.times),
                tuple(self.values),
                tuple(self.lattice_site_coefficients),
            )
        )


class RabiFrequencyAmplitude(BaseModel):
    global_: GlobalField

    class Config:
        allow_population_by_field_name = True
        fields = {"global_": "global"}

    def __hash__(self):
        return hash((RabiFrequencyAmplitude, self.global_))

    def discretize(self, task_capabilities: QuEraCapabilities):
        global_time_resolution = (
            task_capabilities.capabilities.rydberg.global_.time_resolution
        )
        global_value_resolution = (
            task_capabilities.capabilities.rydberg.global_.rabi_frequency_resolution
        )

        return RabiFrequencyAmplitude(
            global_=GlobalField(
                times=discretize_list(self.global_.times, global_time_resolution),
                values=discretize_list(self.global_.values, global_value_resolution),
            )
        )

    def _get_data_source(self):
        # isolate this for binding glyph later
        # required by visualization
        src = {
            "times_amp": [float(i) for i in self.global_.times],
            "values_amp": [float(i) for i in self.global_.values],
            "values_floor_amp": [0.0] * len(self.global_.values),
        }

        return src

    def figure(self, **fig_kwargs):
        return get_task_ir_figure(self, **fig_kwargs)

    def show(self):
        display_task_ir(self)


class RabiFrequencyPhase(BaseModel):
    global_: GlobalField

    class Config:
        allow_population_by_field_name = True
        fields = {"global_": "global"}

    def __hash__(self):
        return hash((RabiFrequencyPhase, self.global_))

    def discretize(self, task_capabilities: QuEraCapabilities):
        global_time_resolution = (
            task_capabilities.capabilities.rydberg.global_.time_resolution
        )
        global_value_resolution = (
            task_capabilities.capabilities.rydberg.global_.phase_resolution
        )

        return RabiFrequencyPhase(
            global_=GlobalField(
                times=discretize_list(self.global_.times, global_time_resolution),
                values=discretize_list(self.global_.values, global_value_resolution),
            )
        )

    def _get_data_source(self):
        # isolate this for binding glyph later
        src = {
            "times_phase": [float(i) for i in self.global_.times],
            "values_phase": [float(i) for i in self.global_.values],
        }

        return src

    def figure(self, **fig_kwargs):
        ## fig_kwargs is for extra tuning when assemble
        ## e.g. calling from QuEraTaskSpecification.figure()
        return get_task_ir_figure(self, **fig_kwargs)

    def show(self):
        # we dont need fig_kwargs when display alone
        display_task_ir(self)


class Detuning(BaseModel):
    global_: GlobalField
    local: Optional[LocalField]

    class Config:
        allow_population_by_field_name = True
        fields = {"global_": "global"}

    def __hash__(self):
        return hash((Detuning, self.global_, self.local))

    def discretize(self, task_capabilities: QuEraCapabilities):
        global_time_resolution = (
            task_capabilities.capabilities.rydberg.global_.time_resolution
        )
        global_value_resolution = (
            task_capabilities.capabilities.rydberg.global_.detuning_resolution
        )
        local_time_resolution = (
            task_capabilities.capabilities.rydberg.local.time_resolution
        )

        if self.local is not None:
            self.local = LocalField(
                times=discretize_list(self.local.times, local_time_resolution),
                values=self.local.values,
                lattice_site_coefficients=self.local.lattice_site_coefficients,
            )

        return Detuning(
            global_=GlobalField(
                times=discretize_list(self.global_.times, global_time_resolution),
                values=discretize_list(self.global_.values, global_value_resolution),
            ),
            local=self.local,
        )

    def _get_data_source(self):
        # isolate this for binding glyph later
        src = {
            "times_detune": [float(i) for i in self.global_.times],
            "values_detune": [float(i) for i in self.global_.values],
            "values_floor_detune": [0.0] * len(self.global_.values),
        }

        return src

    def global_figure(self, **fig_kwargs):
        return get_task_ir_figure(self, **fig_kwargs)

    def show_global(self):
        display_task_ir(self)


class RydbergHamiltonian(BaseModel):
    rabi_frequency_amplitude: RabiFrequencyAmplitude
    rabi_frequency_phase: RabiFrequencyPhase
    detuning: Detuning

    def __hash__(self):
        return hash(
            (
                RydbergHamiltonian,
                self.rabi_frequency_amplitude,
                self.rabi_frequency_phase,
                self.detuning,
            )
        )

    def discretize(self, task_capabilities: QuEraCapabilities):
        return RydbergHamiltonian(
            rabi_frequency_amplitude=self.rabi_frequency_amplitude.discretize(
                task_capabilities
            ),
            rabi_frequency_phase=self.rabi_frequency_phase.discretize(
                task_capabilities
            ),
            detuning=self.detuning.discretize(task_capabilities),
        )


class EffectiveHamiltonian(BaseModel):
    rydberg: RydbergHamiltonian

    def __hash__(self):
        return hash((EffectiveHamiltonian, self.rydberg))

    def discretize(self, task_capabilities: QuEraCapabilities):
        return EffectiveHamiltonian(rydberg=self.rydberg.discretize(task_capabilities))


class Lattice(BaseModel):
    sites: List[Tuple[FloatType, FloatType]]
    filling: List[int]

    def __hash__(self):
        return hash((Lattice, tuple(self.sites), tuple(self.filling)))

    def discretize(self, task_capabilities: QuEraCapabilities):
        position_resolution = (
            task_capabilities.capabilities.lattice.geometry.position_resolution
        )
        return Lattice(
            sites=[discretize_list(site, position_resolution) for site in self.sites],
            filling=self.filling,
        )

    def figure(self, **fig_kwargs):
        ## use ir.Atom_oarrangement's plotting:
        ## covert unit to m -> um
        return get_task_ir_figure(self, **fig_kwargs)

    def show(self):
        display_task_ir(self)


class QuEraTaskSpecification(BaseModel):
    nshots: int
    lattice: Lattice
    effective_hamiltonian: EffectiveHamiltonian

    def __hash__(self):
        return hash(
            (
                QuEraTaskSpecification,
                self.nshots,
                self.lattice,
                self.effective_hamiltonian,
            )
        )

    def discretize(self, task_capabilities: QuEraCapabilities):
        return QuEraTaskSpecification(
            nshots=self.nshots,
            lattice=self.lattice.discretize(task_capabilities),
            effective_hamiltonian=self.effective_hamiltonian.discretize(
                task_capabilities
            ),
        )

    def figure(self):
        return get_task_ir_figure(self)

    def show(self):
        display_task_ir(self)



from bloqade.ir import var, cast, Variable, Literal, start
from bloqade.ir import to_waveform as waveform
from bloqade.serialize import load, save, loads, dumps

from bloqade.factory import (
    get_capabilities,
    piecewise_linear,
    piecewise_constant,
    linear,
    constant,
    rydberg_h,
)
import bloqade.ir as _ir
from bloqade.constants import RB_C6

import importlib.metadata

__version__ = importlib.metadata.version("bloqade")


def tree_depth(depth: int = None):
    """Setting globally maximum depth for tree printing

    If `depth=None`, return current depth.
    If `depth` is provided, setting current depth to `depth`

    Args:
        depth (int, optional): the user specified depth. Defaults to None.

    Returns:
        int: current updated depth
    """
    if depth is not None:
        _ir.tree_print.max_tree_depth = depth
    return _ir.tree_print.max_tree_depth


__all__ = [
    "RB_C6",
    "start",
    "var",
    "cast",
    "Variable",
    "Literal",
    "piecewise_linear",
    "piecewise_constant",
    "linear",
    "constant",
    "set_print_depth",
    "load",
    "save",
    "loads",
    "dumps",
    "rydberg_h",
    "waveform",
    "get_capabilities",
]



from beartype import beartype
from beartype.typing import List, Optional, Union
from bloqade.ir.scalar import Variable
from bloqade.builder.base import Builder
from bloqade.builder.pragmas import Parallelizable
from bloqade.builder.backend import BackendRoute


class Args(Parallelizable, BackendRoute, Builder):
    @beartype
    def __init__(
        self, order: List[Union[str, Variable]], parent: Optional[Builder] = None
    ) -> None:
        super().__init__(parent)
        self._order = tuple([o.name if isinstance(o, Variable) else o for o in order])
